
import os
import zmq
import errno

from ..url  import Url
from ..host import get_hostip
from ..misc import as_list, as_string, find_port, ru_open


# NOTE: this is ignoring `RADICAL_LOG_LVL` on purpose
LOG_ENABLED = os.environ.get('RADICAL_ZMQ_LOG', '0').lower() in ['1', 'true']


# --------------------------------------------------------------------------
#
# zmq will (rightly) barf at interrupted system calls.  We are able to rerun
# those calls.
#
# This is presumably rare, and repeated interrupts increasingly unlikely.
# More than, say, 3 point to races or I/O thrashing
#
# FIXME: how does that behave wrt. timeouts?  We probably should include
#        an explicit timeout parameter.
#
# kudos: https://gist.github.com/minrk/5258909
#
def no_intr(f, *args, **kwargs):

    _max = 3
    cnt  = 0
    while True:
        try:
            return f(*args, **kwargs)

        except zmq.ContextTerminated:
            return None    # connect closed or otherwise became unusable

        except zmq.ZMQError as e:
            if e.errno == errno.EINTR:
                if cnt > _max:
                    raise  # interrupted too often - forward exception
                cnt += 1
                continue   # interrupted, try again
            raise          # some other error condition, raise it


# ------------------------------------------------------------------------------
#
def get_uids(msgs):

    msgs = as_list(msgs)
    try   : return [str(m.get('uid')) for m in msgs]
    except: return []


# ------------------------------------------------------------------------------
#
def prof_bulk(prof, event, msgs, msg=None):

    for uid in get_uids(msgs):
        if uid:
            prof.prof(event, uid=uid)
        else:
            prof.prof(event, msg=msgs[:32])


# ------------------------------------------------------------------------------
#
def get_channel_url(ep_type, channel=None, url=None):
    '''
    For the given endpoint type, ensure that both channel name and endpoint URL
    are known.  If they are not, raise a ValueError exception.

    For a given URL, the channel is derived as path element of that URL
    (leading `/` is stripped).

    For a given channel channel name, the URL is searched in the process
    environment (under uppercase version of `<CHANNEL>_<EPTYPE>_URL`).  If not
    found, the method will look if a config file with the name `<channel>.cfg`
    exists, and if it has a top level entry named `<ep_type>` (lower case).

    Before returning the given or derived channel and url, the method will check
    if both data match (i.e. if the channel name is reflected in the URL)
    '''

    if not channel and not url:
        raise ValueError('need either channel name or URL')

    if not channel:
        # get channel from path element of URL
        # example:
        #   channel `foo`
        #   url     `pubsub://localhost:1234/foo`
        channel = os.path.basename(str(Url(url.path)))

    elif not url:
        # get url from environment (`FOO_PUB_URL`) or config file (`foo.cfg`)

        env_name = '%s_%s_URL' % (channel.upper(), ep_type.upper())
        cfg_name = './%s.cfg'  %  channel.lower()

        if env_name in os.environ:
            url = os.environ[env_name]

        elif os.path.exists(cfg_name):
            with ru_open(cfg_name, 'r') as fin:
                for line in fin.readlines():
                    _ep_type, _url = line.split(':')
                    if _ep_type.strip().upper() == ep_type.upper():
                        url = _url
                        break

    # sanity checks
    if not url:
        raise ValueError('no URL for %s channel %s' % (channel, ep_type))

    if not channel:
        raise ValueError('no %s channel for URL %s' % (ep_type, url))

    if channel.lower() != Url(url).path.lstrip('/').lower():
        raise ValueError('%s channel (%s) / url (%s) mismatch'
                        % (ep_type, channel, url))

    return channel, url


# ------------------------------------------------------------------------------
#
def log_bulk(log, token, msgs):

    if log.num_level > 1:
        # log level `debug_9` disabled
        return

    if not msgs:
        return

    if isinstance(msgs[0], dict) and 'uid' in msgs[0]:
        for msg in msgs:
            log.debug_9("%s: %s [%s]", token, msg['uid'], msg.get('state'))

    else:
        for msg in msgs:
            log.debug_9("%s: %s", token, str(msg)[0:32])


# ------------------------------------------------------------------------------
#
def sock_connect(sock, url, hop=None):

    if not hop:
        hop = os.environ.get('RADICAL_SSH_HOP')

    if hop:
        from zmq import ssh
        ssh.tunnel_connection(sock, url, hop)

    else:
        sock.connect(url)


# ------------------------------------------------------------------------------
#
def zmq_bind(sock):

    while True:
        port = find_port()
        try:
            sock.bind('tcp://*:%s' % port)
            addr      = Url(as_string(sock.getsockopt(zmq.LAST_ENDPOINT)))
            addr.host = get_hostip()
            return addr
        except:
            pass

    raise RuntimeError('could not bind to any port')


# ------------------------------------------------------------------------------





import zmq
import threading as mt

from ..serialize import to_msgpack, from_msgpack
from ..logger    import Logger

from .utils import zmq_bind

MODE_PUSH = 'push'
MODE_PULL = 'pull'


# ------------------------------------------------------------------------------
#
class Pipe(object):
    '''
    The `Pipe` class provides simple and direct n-to-m connectivity without
    load balancing or caching.  The class is expected to be used in the
    following way:

      - *one* endpoint (either push or pull) establishes the pipe:
        p = Pipe()
        p.connect_push()
        print(p.url)

      - all other endpoints (either push or pull) connect to the same pipe by
        specifying the URL from the first EP:
        p1 = Pipe()
        p1.connect_pull(url)

      - data should only be sent once at least one receiving EP is established
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, mode, url=None, log=None) -> None:
        '''
        Create a `Pipe` instance which can be used for either sending (`put()`)
        or receiving (`get()` / `get_nowait()`) data. according to the specified
        mode (`MODE_PUSH` or `MODE_PULL`).

        An URL can be specified for one end of the pipe - that end will then be
        in listening mode.  The other end of the pipe MUST use the connection
        URL provided by the listening end (`Pipe.url`).
        '''

        self._context = zmq.Context.instance()
        self._mode    = mode
        self._url     = None
        self._log     = log
        self._sock    = None
        self._poller  = None
        self._cbs     = list()
        self._thread  = None
        self._term    = mt.Event()

        if mode == MODE_PUSH:
            self._connect_push(url)

        elif mode == MODE_PULL:
            self._connect_pull(url)

        else:
            raise ValueError('unsupported pipe mode [%s]' % mode)

        if not self._log:

            self._log = Logger('radical.utils.pipe')


    # --------------------------------------------------------------------------
    #
    @property
    def url(self):
        return self._url


    # --------------------------------------------------------------------------
    #
    def _connect_push(self, url):
        '''
        Establish this pipe instance as sending endpoint.
        '''

        assert self._mode == MODE_PUSH

        if self._sock:
            raise RuntimeError('already connected at %s' % self._url)

        self._sock = self._context.socket(zmq.PUSH)

        if url:
            self._sock.connect(url)
            self._url = url
        else:
            self._url = zmq_bind(self._sock)

        self._url = self._sock.getsockopt(zmq.LAST_ENDPOINT)


    # --------------------------------------------------------------------------
    #
    def _connect_pull(self, url):
        '''
        Establish this Pipe as receiving endpoint.
        '''

        assert self._mode == MODE_PULL

        if self._sock:
            raise RuntimeError('already connected at %s' % self._url)

        self._sock = self._context.socket(zmq.PULL)

        if url:
            self._sock.connect(url)
            self._url = url
        else:
            self._url = zmq_bind(self._sock)

        self._poller = zmq.Poller()
        self._poller.register(self._sock, zmq.POLLIN)


    # --------------------------------------------------------------------------
    #
    def get(self):
        '''
        Receive a message.  This call blocks until a message is available.
        '''

        assert self._mode == MODE_PULL
        assert not self._cbs

        return from_msgpack(self._sock.recv())


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, timeout: float = 0):
        '''
        Receive a message.  This call blocks for `timeout` seconds
        until a message is available.  If no message is available after timeout,
        `None` is returned.
        '''

        assert self._mode == MODE_PULL
        assert not self._cbs

        # zmq timeouts are in milliseconds
        socks = dict(self._poller.poll(timeout=int(timeout * 1000)))

        if self._sock in socks:
            return from_msgpack(self._sock.recv())


    # --------------------------------------------------------------------------
    #
    def _listener(self):
        '''
        Listen for incoming messages, and call registered callbacks.
        '''

        while not self._term.is_set():

            socks = dict(self._poller.poll(timeout=10))

            if self._sock in socks:
                msg = from_msgpack(self._sock.recv())

                for cb in self._cbs:
                    try:
                        cb(msg)
                    except:
                        self._log.exception('callback failed')


    # --------------------------------------------------------------------------
    #
    def register_cb(self, cb):
        '''
        Register a callback for incoming messages.  The callback will be called
        with the message as argument.

        Only a pipe in pull mode can have callbacks registered.  Note that once
        a callback is registered, the `get()` and `get_nowait()` methods must
        not be used anymore.
        '''

        assert self._mode == MODE_PULL

        self._cbs.append(cb)

        if not self._thread:
            self._thread = mt.Thread(target=self._listener)
            self._thread.daemon = True
            self._thread.start()


    # --------------------------------------------------------------------------
    #
    def unregister_cb(self, cb):
        '''
        Unregister a callback.  If no callback remains registered, the listener
        thread will be stopped.
        '''

        assert self._mode == MODE_PULL
        assert cb in self._cbs

        self._cbs.remove(cb)

        if not self._cbs:
            self._stop_listener()


    # --------------------------------------------------------------------------
    #
    def put(self, msg):
        '''
        Send a message - if receiving endpoints are connected, exactly one of
        them will be able to receive that message.
        '''

        assert self._mode == MODE_PUSH
        self._sock.send(to_msgpack(msg))


    # --------------------------------------------------------------------------
    #
    def _stop_listener(self):

        if self._thread:
            self._term.set()
            self._thread.join()
            self._term.clear()
            self._thread = None


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._stop_listener()



# ------------------------------------------------------------------------------





import os
import sys
import time
import errno
import socket
import tarfile
import datetime
import tempfile
import itertools

from typing import List, Tuple, Union, Any

from .         import url       as ruu
from .ru_regex import ReString

# ------------------------------------------------------------------------------
#
# globals
#
_RU_stdout = None
_RU_stderr = None
_RU_except = None
_RU_exit   = None


# ------------------------------------------------------------------------------
#
_RU_stdout = None
_RU_stderr = None
_RU_except = None
_RU_exit   = None


# ------------------------------------------------------------------------------
#
def parse_file_staging_directives(directives):
    '''
    staging directives

       [local_path] [operator] [remote_path]

    local path:
        * interpreted as relative to the application's working directory
        * must point to local storage (localhost)

    remote path
        * interpreted as relative to the job's working directory

    operator :
        * >  : stage to remote target, overwrite if exists
        * >> : stage to remote target, append    if exists
        * <  : stage to local  target, overwrite if exists
        * << : stage to local  target, append    if exists

    This method returns a tuple [src, tgt, op] for each given directive.  This
    parsing is backward compatible with the simple staging directives used
    previously -- any strings which do not contain staging operators will be
    interpreted as simple paths (identical for src and tgt), operation is set to
    '=', which must be interpreted in the caller context.
    '''

    bulk = True
    if  not isinstance(directives, list):
        bulk       = False
        directives = [directives]

    ret = list()

    for directive in directives:

        if  not is_string(directive):
            raise TypeError("file staging directives muct by of type string, "
                            "not %s" % type(directive))

        rs = ReString(directive)

        if  rs // r'^(?P<one>.+?)\s*(?P<op><|<<|>|>>)\s*(?P<two>.+)$':
            res = rs.get()
            ret.append([res['one'], res['two'], res['op']])

        else:
            ret.append([directive, directive, '='])

    if  bulk: return ret
    else    : return ret[0]


# ------------------------------------------------------------------------------
#
def time_stamp(spec):

    if  isinstance(spec, int) or \
        isinstance(spec, float)  :

        return datetime.datetime.utcfromtimestamp(spec)

    return spec


# ------------------------------------------------------------------------------
#
def time_diff(dt_abs, dt_stamp):
    '''
    return the time difference bewteen  two datetime
    objects in seconds (incl. fractions).  Exceptions (like on improper data
    types) fall through.
    '''

    delta = dt_stamp - dt_abs

    # make it easy to use seconds since epoch instead of datetime objects
    if  isinstance(delta, int) or \
        isinstance(delta, float)  :
        return delta

    if  not isinstance(delta, datetime.timedelta):
        raise TypeError("difference between '%s' and '%s' is not a .timedelta"
                     % (type(dt_abs), type(dt_stamp)))

    # get seconds as float
    seconds = delta.seconds + delta.microseconds / 1E6
    return seconds


# ------------------------------------------------------------------------------
#
def all_pairs(iterable, n):
    '''
    [ABCD] -> [AB], [AC], [AD], [BC], [BD], [CD]
    '''

    return list(itertools.combinations(iterable, n))


# ------------------------------------------------------------------------------
#
def cluster_list(iterable, n):
    '''
    s -> [ s0,  s1,    s2,    ... sn-1  ],
         [ sn,  sn+1,  sn+2,  ... s2n-1 ],
         [ s2n, s2n+1, s2n+2, ... s3n-1 ],
         ...
    '''

    return zip(*[iter(iterable)] * n)


# ------------------------------------------------------------------------------
# From https://docs.python.org/release/2.3.5/lib/itertools-example.html
#
def window(seq, n=2):
    '''
    Returns a sliding window (of width n) over data from the iterable"
    s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
    '''

    it = iter(seq)
    result = tuple(itertools.islice(it, n))

    if len(result) == n:
        yield result

    for elem in it:
        result = result[1:] + (elem,)
        yield result


# ------------------------------------------------------------------------------
#
def round_to_base(value: float, base: int = 1) -> int:
    '''
    This method expects an integer or float value, and will round it to any
    given integer base.  For example:

      1.5, 2 -> 2
      3.5, 2 -> 4
      4.5, 2 -> 4

      11.5, 20 -> 20
      23.5, 20 -> 20
      34.5, 20 -> 40

    The default base is '1'.
    '''

    return int(base * round(float(value) / base))


# ------------------------------------------------------------------------------
#
def round_upper_bound(value: Union[int, float]) -> int:
    '''
    This method expects an integer or float value, and will return an integer
    upper bound suitable for example to define plot ranges.  The upper bound is
    the smallest value larger than the input value which is a multiple of 1,
    2 or 5 times the order of magnitude (10**x) of the value.
    '''

    bound = 0
    order = 0
    check = [1, 2, 5]

    while True:

        for c in check:

            bound = c * (10**order)

            if value < bound:
                return bound

        order += 1


# ------------------------------------------------------------------------------
#
def is_tuple(data: Any) -> bool:
    '''
    return True if given data are a `tuple`, `False` otherwise
    '''

    return isinstance(data, tuple)


# ------------------------------------------------------------------------------
#
def as_tuple(data: Any) -> Tuple[Any]:
    '''
    return non-tuple data into a tuple.
    '''

    if   data is None  : return tuple()
    elif is_tuple(data): return data
    else               : return (data, )


# ------------------------------------------------------------------------------
#
def is_list(data: Any) -> bool:
    '''
    return True if given data are a `list`, `False` otherwise
    '''

    return isinstance(data, list)


# ------------------------------------------------------------------------------
#
def as_list(data: Any) -> List[Any]:
    '''
    return non-list data into a list.
    '''

    if   data is None : return []
    elif is_list(data): return data
    else              : return [data]


# ------------------------------------------------------------------------------
#
def to_type(data: Any) -> Union[str, int, float, Any]:

    if not isinstance(data, str):
        return data

    try   : return(int(data))
    except: pass

    try   : return(float(data))
    except: pass

    return data


# ------------------------------------------------------------------------------
#
def is_seq(data: Any) -> bool:
    '''
    tests if the given data is a sequence (but not a string)
    '''
    return hasattr(data, '__iter__') and not is_string(data)


# ------------------------------------------------------------------------------
#
def is_string(data: Any) -> bool:
    '''
    tests if the given data are a `string` type
    '''
    return isinstance(data, str)


# ------------------------------------------------------------------------------
#
def as_string(data: Any) -> Any:
    '''
    Make a best-effort attempt to convert bytes to strings.  Iterate through
    lists and dicts, but leave all other datatypes alone.
    '''

    if isinstance(data, dict):
        return {as_string(k): as_string(v) for k,v in data.items()}

    elif isinstance(data, list):
        return [as_string(e) for e in data]

    elif isinstance(data, bytes):
        return bytes.decode(data, 'utf-8')

    elif isinstance(data, ruu.Url):
        return str(data)

    else:
        return data


# ------------------------------------------------------------------------------
#
def is_bytes(data: Any) -> bool:
    '''
    checks if the given data are of types `bytes` or `bytearray`
    '''
    return isinstance(data, (bytes, bytearray))


# ------------------------------------------------------------------------------
# thanks to
# http://stackoverflow.com/questions/956867/#13105359
def as_bytes(data: Any) -> Any:
    '''
    Make a best-effort attempt to convert strings to bytes.  Iterate through
    lists and dicts, but leave all other datatypes alone.
    '''

    if isinstance(data, dict):
        return {as_bytes(k): as_bytes(v) for k,v in data.items()}

    elif isinstance(data, list):
        return [as_bytes(e) for e in data]

    elif isinstance(data, str):
        return str.encode(data, 'utf-8')

    else:
        return data


# ------------------------------------------------------------------------------
#
def watch_condition(cond, target=None, timeout=None, interval=0.1):
    '''
    Watch a given condition (a callable) until it returns the target value, and
    return that value.  Stop watching on timeout, in that case return None.  The
    condition is tested approximately every 'interval' seconds.
    '''

    start = time.time()
    while True:
        ret = cond()
        if ret == target:
            return ret
        if timeout and time.time() > start + timeout:
            return None
        time.sleep(interval)


# ------------------------------------------------------------------------------
#
def name2env(name):
    '''
    convert a name of the for 'radical.pilot' to an env vare base named
    'RADICAL_PILOT'.
    '''

    return name.replace('.', '_').upper()


# ------------------------------------------------------------------------------
#
def get_env_ns(key, ns, default=None):
    '''
    get an environment setting within a namespace.  For example.

        get_env_ns('verbose', 'radical.pilot.tmgr'),

    will return the value of the first found env variable from the following
    sequence:

        RADICAL_PILOT_TMGR_LOG_LVL
        RADICAL_PILOT_LOG_LVL
        RADICAL_LOG_LVL

    or 'None' if none of the above is set.  The given `name` and `key` are
    converted to upper case, dots are replaced by underscores.

    Note that an environment variable set with

        export RADICAL_LOG_LVL=

    (ie. without an explicit, non-empty value) will be returned as an empty
    string.
    '''

    ns     = name2env(ns)
    key    = name2env(key)
    base   = ''
    checks = list()
    for elem in ns.split('_'):
        base += elem + '_'
        check = base + key
        checks.append(check)

    for check in reversed(checks):
        if check in os.environ:
            val = os.environ[check]
            return val

    return default


# ------------------------------------------------------------------------------
#
def expand_env(data, env=None, ignore_missing=True):
    '''
    Expand the given data with environment variables from `os.environ`.
    If `env` is provided, use that dictionary for expansion instead.

    `data` can be one of three types:

      - dictionary: `expand_env` is applied to all *values* of the dictionary
      - sequence  : `expand_env` is applied to all elements of the sequence
      - string    : `expand_env` is applied to the string itself


    The method will alter dictionaries and iterables in place, but will return
    a copy of scalar strings, as it seems to be custom in Python.  Other data
    types are silently ignored and not altered.

    The replacement in strings is performed for the following variable specs:

        assume  `export BAR=bar BIZ=biz`:

            ${BAR}     : foo_${BAR}_baz      -> foo_bar_baz
            ${BAR:buz} : foo_${BAR:buz}_baz  -> foo_bar_baz
            ${BAR:$BUZ}: foo_${BAR:$BIZ}_baz -> foo_biz_baz

        assume `unset BAR; export BIZ=biz`, `ignore_missing=True`

            ${BAR}     : foo_${BAR}_baz      -> foo__baz
            ${BAR:buz} : foo_${BAR:buz}_baz  -> foo_buz_baz
            ${BAR:$BUZ}: foo_${BAR:$BIZ}_baz -> foo_biz_baz

        assume `unset BAR; export BIZ=biz`, `ignore_missing=False`

            ${BAR}     : foo_${BAR}_baz      -> ValueError('cannot expand $BAR')
            ${BAR:buz} : foo_${BAR:buz}_baz  -> foo_buz_baz
            ${BAR:$BIZ}: foo_${BAR:$BIZ}_baz -> foo_biz_baz

    The method will also opportunistically convert strings to integers or
    floats if they are formatted that way and contain no other characters.
    '''

    # no data: None, empty dict / sequence / string
    if not data:
        return data

    # dict type
    elif isinstance(data, dict):

        for k,v in data.items():
            data[k] = expand_env(v, env, ignore_missing)
        return data

    # sequence types: list, set, tuple - but not string
    elif is_seq(data):

        for idx, elem in enumerate(data):
            data[idx] = expand_env(elem, env, ignore_missing)
        return data

    # all other non-string types are left alone
    elif not is_string(data):
        return data

    if '$' not in data:
        # nothing to expand
        return data

    # fall back to process env if no other expansion dict is specified
    if not env:
        env = os.environ

    # strings are not expanded in place - create a new one to fill.
    # iterate over the orginial string as long as there is something to expand
    ret = ''
    while data:

        data = ReString(data)

        # idea     :   pre     ${  Vari_ABLE            : val     } post
        # captures :  (   )(?    (                  )(?  (     ))  )(  )
        # indexes  :  1          2                       3          4
        with data // r'(.*?)(?:\${([a-zA-Z][a-zA-Z0-9_-]+)(?::([^}]+))?})(.*)' \
            as res:

            if not res:
                ret += data
                break

            pre  = res[0]
            key  = res[1]
            val  = res[2]
            post = res[3]

            if not ignore_missing and key not in env:
                raise ValueError('cannot expand $%s' % key)

            if pre  is None: pre  = ''
            if val  is None: val  = ''
            if post is None: post = ''

            # support env expansion of val, as in
            #   LOGDIR : "${RCT_LOGDIR:$PWD}"
            if val.startswith('$'):
                val = env.get(val[1:], '')

            val = env.get(key, val)

            if key and not pre and not post and not val:
                # we had something to expand, and that expansion is all there is
                # in the key, and the expand failed - then the result it not an
                # empty string but None
                return None

            ret += pre
            ret += val

            data = ReString(post)

    # attempt string-to-type conversion (int and float detection only)
    return to_type(ret)


# ------------------------------------------------------------------------------
#
def get_size(obj, seen=None, strict=False):

    size   = sys.getsizeof(obj)
    obj_id = id(obj)

    if strict:
        # perform recursion checks
        if seen is None:
            seen = set()
        if obj_id in seen:
            return 0
        seen.add(obj_id)

    if isinstance(obj, dict):
        size += sum([get_size(v, seen, strict) for v in obj.values()])
        size += sum([get_size(k, seen, strict) for k in obj.keys()])

    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen, strict)

    elif hasattr(obj, '__iter__') and \
        not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen, strict) for i in obj])

    return size


# ------------------------------------------------------------------------------
#
def dockerized():

    if os.path.exists('/.dockerenv'):
        return True
    return False


# ------------------------------------------------------------------------------
#
def get_radical_base(module=None):
    '''
    Several parts of the RCT stack store state on the file system.  This should
    usually be under `$HOME/.radical` - but that location is not always
    available or desireable.  We interpret the env variable `RADICAL_BASE`,
    and fall back to `pwd` if neither that nor `$HOME` exists.

    The optional `module` parameter will result in the respective subdir name to
    be appended.  The resulting dir is created (if it does not exist), and the
    name is returned.  Any `.` (dot) characters in `module` are replaced by
    slashes.  Leading `radical/` element is removed.
    '''

    return get_base(ns='radical', module=module)


# ------------------------------------------------------------------------------
#
def get_base(ns, module=None):
    '''
    A generic version of `get_radical_base` which queries the base for any
    namespace `ns`
    '''

    ns_low = ns.lower()
    ns_up  = ns.upper()

    if module:
        module = module.replace('.', '/')
        if module.startswith('%s/' % ns_low):
            module = module[len(ns_low) + 1:]

    base = os.environ.get("%s_BASE" % ns_up)

    if not base:
        base  = os.environ.get("HOME")

    if not base:
        base  = os.environ.get("PWD")

    if not base:
        base  = os.getcwd()

    if module: base += '/.%s/%s/' % (ns_low, module)
    else     : base += '/.%s/'    %  ns_low

    rec_makedir(base)

    return base


# ------------------------------------------------------------------------------
#
def rec_makedir(target):
    '''
    recursive makedir which ignores errors if dir already exists
    '''

    try:
        os.makedirs(target)

    except OSError as e:

        # ignore failure on existing directory - otherwise raise
        if e.errno == errno.EEXIST and os.path.isdir(os.path.dirname(target)):
            return

        raise


# ------------------------------------------------------------------------------
#
def mktar(tarname, fnames=None, data=None):
    '''
    Create a tarfile at the given `tarname`, and pack all files given in
    `fnames` into it, and also pack any `data` blobs.

    `fnames` is expected to be list, where each element is either a string
    pointing to a file to be added under that name, or a tuple where the first
    element points again to the file to be packed, and the second element
    specifies the name under which the file should be packed into the archive.
    And OSError will be raised if the file does not exist.

    `data` is expected to be a list of tuples, where the first element is a set
    of bytes comprising the data to be written into the archive, and the second
    element again specifies the name of the tarred file.

    Note that this method always create bzip'ed tarfiles, but will never change
    the `tarname` to reflect that.
    '''

    tar = tarfile.open(tarname, "w:bz2")
    if fnames:
        for element in fnames:
            if isinstance(str, element):
                tar.add(element)
            else:
                src, tgt = element
                tar.add(src, arcname=tgt)

    if data:
        for fname, fdata in data:
            tmp_name, tmp_fd = tempfile.mkstemp()
            tmp_fd.write(fdata)
            tmp_fd.close()
            tar.add(tmp_name, fname)
            os.unlink(tmp_name)

    tar.close()


# ------------------------------------------------------------------------------
#
def noop(*args, **kwargs):
    return None


# ------------------------------------------------------------------------------
#
def script_2_func(fpath):
    '''
    This method accepts a single parameter `fpath` which is expected to point to
    a file containing a self-sufficient Python script.  The script will be read
    and stored, and a function handle will be returned which, upon calling, will
    run that script in the current Python interpreter.  It will be ensured that
    `__name__` is set to `__main__`, and that any arguments passed to the
    callable are passed on as `sys.argv`.  A single list argument is also
    allowed which is interpreted as argument list.

    Example:

        my_func = ru.script_2_func('/tmp/my_script.py')
        my_func('-f', 'foo', '-b', 'bar')

    NOTE: calling the returned function handle will change `sys.argv` for the
          current Python interpreter.
    '''

    prefix  = []
    postfix = []

    with ru_open(fpath, 'r') as fin:
        code_lines = fin.readlines()

    # ensure that 'if __name__ == '__main__' works
    prefix  += '__name__ = "__main__"\n\n'

    # capture globals
    prefix.append('global _RU_stdout\n')
    prefix.append('global _RU_stderr\n')
    prefix.append('global _RU_except\n')
    prefix.append('global _RU_exit\n')

    # redirect stdout and stderr to be captured in global string vars
    prefix.append('from io import StringIO\n')
    prefix.append('_RU_oldout  = sys.stdout\n')
    prefix.append('_RU_olderr  = sys.stderr\n')
    prefix.append('sys.stdout  = _RU_stdout = StringIO()\n\n')
    prefix.append('sys.stderr  = _RU_stderr = StringIO()\n\n')

    # wrap the code in a try/exec to capture exit codes and failures
    prefix.append('_RU_except = None\n')
    prefix.append('try:\n')

    postfix.append('except Exception as e:\n')
    postfix.append('    _RU_except = str(e)\n')
    postfix.append('    _RU_exit   = 1\n')
    postfix.append('except SystemExit as e:\n')
    postfix.append('    _RU_except = "SystemExit"\n')
    postfix.append('    _RU_exit   = e.code\n')

    postfix.append('sys.stdout = _RU_oldout\n')
    postfix.append('sys.stderr = _RU_olderr\n')

    # examine _RU_stdout.getvalue()

    # indentation of all code lines which go into the try/except
    code_lines = ['    ' + line for line in code_lines]

    # create closure which can be used to call the code
    def ret(*argv):

        if len(argv) == 1 and isinstance(argv[0], list):
            argv = list(argv[0])

        # if args are given, ensure that sys.argv gets set
        if argv:
            # indent - this is within try/catch
            prefix.append('    import sys\n')
            prefix.append('    sys.argv = ["dummy"] + %s\n' % str(list(argv)))

        global _RU_stdout
        global _RU_stderr
        global _RU_except
        global _RU_exit

        _RU_stdout = None
        _RU_stderr = None
        _RU_except = None
        _RU_exit   = None

        tmp_code = ''.join(prefix)   \
                 + ''.join(code_lines) \
                 + ''.join(postfix)

        # exec the resulting code, ensure to pass globals
        exec(tmp_code, globals())                        # pylint: disable=W0122

        return _RU_stdout.getvalue(), _RU_stderr.getvalue(), \
               _RU_except, _RU_exit

    # return that new callable
    return ret

  # # --------------------------------------------------------------------------
  #
  # import importlib
  #
  # loader = importlib.machinery.SourceFileLoader('__main__', fpath)
  # spec   = importlib.util.spec_from_loader(loader.name, loader)
  # mod    = importlib.util.module_from_spec(spec)
  # code   = loader.get_code(mod.__name__)
  #
  # def ret(*argv):
  #     args = list(argv)
  #     if len(args) == 1 and isinstance(argv[0], list):
  #         args = list(argv[0])
  #     sys.argv = ['dummy'] + args
  #     loader.exec_module(mod)
  #
  # return ret


# ------------------------------------------------------------------------------
#
def ru_open(*args, **kwargs):
    '''
    ensure that we use UTF8 consistently throughout the stack
    '''

    if 'encoding' not in kwargs:
        kwargs['encoding'] = 'utf8'

    return open(*args, **kwargs)


# ------------------------------------------------------------------------------
#
def find_port(port_min=10000, port_max=65535):
    '''
    Find a free port in the given range.  The range defaults to 10000-65535.
    Returns `None` if no free port could be found.
    '''

    for port in range(port_min, port_max):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.bind(('', port))
            return port

        except socket.error:
            pass

        finally:
            sock.close()


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

import zmq
import time

import threading as mt

from typing      import Optional

from ..atfork    import atfork
from ..config    import Config
from ..ids       import generate_id, ID_CUSTOM
from ..misc      import as_string, as_bytes, as_list, noop
from ..logger    import Logger
from ..profile   import Profiler
from ..serialize import to_msgpack, from_msgpack

from .bridge     import Bridge
from .utils      import zmq_bind, no_intr, log_bulk, LOG_ENABLED


# ------------------------------------------------------------------------------
#
_LINGER_TIMEOUT  =   250  # ms to linger after close
_HIGH_WATER_MARK =     0  # number of messages to buffer before dropping
                          # 0:  infinite


# ------------------------------------------------------------------------------
#
def _atfork_child():
    for subscriber in Subscriber._instances:
        subscriber._callbacks = list()                                    # noqa


atfork(noop, noop, _atfork_child)


# ------------------------------------------------------------------------------
#
class PubSub(Bridge):

    # --------------------------------------------------------------------------
    #
    def __init__(self, channel: str, cfg: Optional[dict] = None, log=None):

        if cfg:
            # create deep copy
            cfg = Config(cfg=cfg)
        else:
            cfg = Config()

        if not cfg.uid:
            cfg.uid = generate_id('%s.bridge.%%(counter)04d' % channel,
                                  ID_CUSTOM)

        super().__init__(cfg, log=log)


    # --------------------------------------------------------------------------
    #
    # protocol independent addr query
    @property
    def type_in(self):
        return 'pub'

    @property
    def type_out(self):
        return 'sub'

    @property
    def addr_in(self):
        return self.addr_pub

    @property
    def addr_out(self):
        return self.addr_sub

    # protocol dependent addr query
    @property
    def addr_pub(self):
        return self._addr_pub

    @property
    def addr_sub(self):
        return self._addr_sub


    # --------------------------------------------------------------------------
    #
    def _bridge_initialize(self):

        self._log.info('initialize bridge %s', self._uid)

        self._lock        = mt.Lock()

        self._ctx         = zmq.Context.instance()  # rely on GC for destruction
        self._xpub        = self._ctx.socket(zmq.XSUB)
        self._xpub.linger = _LINGER_TIMEOUT
        self._xpub.hwm    = _HIGH_WATER_MARK
        self._addr_pub    = zmq_bind(self._xpub)

        self._xsub        = self._ctx.socket(zmq.XPUB)
        self._xsub.linger = _LINGER_TIMEOUT
        self._xsub.hwm    = _HIGH_WATER_MARK
        self._addr_sub    = zmq_bind(self._xsub)

        self._log.info('bridge pub on  %s: %s', self._uid, self._addr_pub)
        self._log.info('       sub on  %s: %s', self._uid, self._addr_sub)

        # make sure bind is active
        time.sleep(0.01)

        # start polling for messages
        self._poll = zmq.Poller()
        self._poll.register(self._xpub, zmq.POLLIN)
        self._poll.register(self._xsub, zmq.POLLIN)


    # --------------------------------------------------------------------------
    #
    def _bridge_work(self):

        # we could use a zmq proxy - but we rather code it directly to have
        # proper logging, timing, etc.  But the code for the proxy would be:
        #
        #     zmq.proxy(socket_pub, socket_sub)
        #
        # That's the equivalent of the code below.

        while not self._term.is_set():

            # timeout in ms
            socks = dict(self._poll.poll(timeout=10))

            if self._xsub in socks:

                # if the sub socket signals a message, it's likely
                # a topic subscription.  Forward that to the pub
                # channel, so the bridge subscribes for the respective
                # message topic.
                msg = self._xsub.recv()
                self._xpub.send(msg)

                self._prof.prof('subscribe', uid=self._uid, msg=msg)
                log_bulk(self._log, '~~1 %s' % self.uid, [msg])


            if self._xpub in socks:

                # if the pub socket signals a message, get the message
                # and forward it to the sub channel, no questions asked.
                msg = self._xpub.recv()
                self._xsub.send(msg)

              # self._prof.prof('msg_fwd', uid=self._uid, msg=msg)
                log_bulk(self._log, '<> %s' % self.uid, [msg])


# ------------------------------------------------------------------------------
#
class Publisher(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, channel, url=None, log=None, prof=None, path=None):

        self._channel  = channel
        self._url      = as_string(url)
        self._log      = log
        self._prof     = prof
        self._lock     = mt.Lock()

        # FIXME: no uid ns
        self._uid = generate_id('%s.pub.%s' % (self._channel,
                                               '%(counter)04d'), ID_CUSTOM)

        if not self._url:
            self._url = Bridge.get_config(channel, path).pub

        if not log:
            if LOG_ENABLED: level = 'DEBUG_9'
            else          : level = 'ERROR'
            self._log = Logger(name=self._uid, ns='radical.utils.zmq',
                               level=level, path=path)

        if not prof:
            self._prof = Profiler(name=self._uid, ns='radical.utils.zmq',
                                  path=path)
            self._prof.disable()

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()

        self._log.info('connect pub to %s: %s', self._channel, self._url)

        self._ctx           = zmq.Context.instance()  # rely on GC for destruction
        self._socket        = self._ctx.socket(zmq.PUB)
        self._socket.linger = _LINGER_TIMEOUT
        self._socket.hwm    = _HIGH_WATER_MARK
        self._socket.connect(self._url)

        time.sleep(0.01)


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def url(self):
        return self._url

    @property
    def channel(self):
        return self._channel


    # --------------------------------------------------------------------------
    #
    def put(self, topic, msg):

        assert isinstance(topic, str), 'invalid topic type'

        self._log.debug_9('put %s : %s: %s', topic, self.channel, msg)
      # self._log.debug_9('put %s: %s', msg, get_stacktrace())
      # self._prof.prof('put', uid=self._uid, msg=msg)
        log_bulk(self._log, '-> %s' % topic, [msg])

        btopic = as_bytes(topic.replace(' ', '_'))
        bmsg   = to_msgpack(msg)
        data   = btopic + b' ' + bmsg

        self._socket.send(data)


# ------------------------------------------------------------------------------
#
class Subscriber(object):

    # We need to clean out some data structures on fork to avoid invalid sockets
    # and deadlock.  For that purpose we keep a list of Subscriber instances
    # around
    _instances = list()


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _get_nowait(socket, lock, timeout, log, prof):

        # FIXME: add logging

        if socket.poll(flags=zmq.POLLIN, timeout=timeout):

            data        = no_intr(socket.recv, flags=zmq.NOBLOCK)
            topic, bmsg = data.split(b' ', 1)
            msg         = from_msgpack(bmsg)

            log.debug_9(' <- %s: %s', topic, msg)

            return [as_string(topic), as_string(msg)]

        return None, None


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _listener(sock, lock, term, callbacks, log, prof):

        try:
            while not term.is_set():

                # this list is dynamic
                topic, msg = Subscriber._get_nowait(sock, lock, 500, log, prof)

                log.debug_9(' <- %s: %s', topic, msg)

                if topic:
                    for cb, _lock in callbacks:
                      # prof.prof('call_cb', uid=uid, msg=cb.__name__)
                        try:
                            if _lock:
                                with _lock:
                                    cb(topic, msg)
                            else:
                                cb(topic, msg)
                        except SystemExit:
                            log.info('callback called sys.exit')
                            term.set()
                            break

                        except:
                            log.exception('callback error')
        except:
            log.exception('listener died')


    # --------------------------------------------------------------------------
    #
    def __init__(self, channel, url=None, topic=None, cb=None,
                                log=None, prof=None, path=None):
        '''
        If a `topic` is given, the channel will subscribe to that topic
        immediately.

        When a callback `cb` is specified, then the Subscriber c'tor will spawn
        a separate thread which continues to listen on the channel, and the cb
        is invoked on any incoming message.  The topic will be the first, the
        message will be the second argument to the cb.
        '''

        Subscriber._instances.append(self)

        self._channel   = channel
        self._url       = as_string(url)
        self._topics    = as_list(topic)
        self._log       = log
        self._prof      = prof

        self._lock      = mt.Lock()
        self._term      = mt.Event()
        self._callbacks = list()
        self._thread    = None
        self._uid       = generate_id('%s.sub.%s' % (self._channel,
                                                    '%(counter)04d'), ID_CUSTOM)

        if not self._topics:
            self._topics = []

        if not self._url:
            self._url = Bridge.get_config(channel, path).sub

        if not self._url:
            raise ValueError('no contact url specified, no config found')

        if not self._log:
            if LOG_ENABLED: level = 'DEBUG_9'
            else          : level = 'ERROR'
            self._log = Logger(name=self._uid, ns='radical.utils.zmq',
                               level=level)

        if not self._prof:
            self._prof = Profiler(name=self._uid, ns='radical.utils.zmq')
            self._prof.disable()

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()

        self._log.info('connect sub to %s: %s', self._channel, self._url)

        self._ctx         = zmq.Context.instance()  # rely on GC for destruction
        self._sock        = self._ctx.socket(zmq.SUB)
        self._sock.linger = _LINGER_TIMEOUT
        self._sock.hwm    = _HIGH_WATER_MARK
        self._sock.connect(self._url)

        time.sleep(0.01)

        # only allow `get()` and `get_nowait()`
        self._interactive = True

        for topic in self._topics:
            self.subscribe(topic, cb)


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def url(self):
        return self._url

    @property
    def channel(self):
        return self._channel


    # --------------------------------------------------------------------------
    #
    def _start_listener(self):

        # only start if needed
        with self._lock:

            if self._thread:
                return

            lock      = self._lock
            term      = self._term
            callbacks = self._callbacks

            self._log.info('start listener for %s', self._channel)

            self._thread = mt.Thread(target=Subscriber._listener,
                                     args=[self._sock, lock, term, callbacks,
                                      self._log, self._prof])
            self._thread.daemon = True
            self._thread.start()


    # --------------------------------------------------------------------------
    #
    def _stop_listener(self, force=False):

        # only stop listener if no callbacks remain registered (unless forced)
        if force or not self._callbacks:
            with self._lock:
                if  self._thread:
                    self._term.set()
                    self._thread.join()
                    self._term.clear()
                    self._thread = None


    # --------------------------------------------------------------------------
    #
    def subscribe(self, topic, cb=None, lock=None):

        # if we need to serve callbacks, then open a thread to watch the socket
        # and register the callbacks.  If a thread is already runnning on that
        # channel, just register the callback.
        #
        # Note that once a thread is watching a socket, we cannot allow to use
        # `get()` and `get_nowait()` anymore, as those will interfere with the
        # thread consuming the messages,
        #
        # The given lock (if any) is used to shield concurrent cb invokations.

        if cb:
            self._interactive = False
            self._start_listener()
            self._callbacks.append([cb, lock])

        topic = str(topic).replace(' ', '_')
        log_bulk(self._log, '~~2 %s' % topic, [topic])

        with self._lock:
            self._log.debug_9('subscribe for %s', topic)
            no_intr(self._sock.setsockopt, zmq.SUBSCRIBE, as_bytes(topic))

        if topic not in self._topics:
            self._topics.append(topic)


    # --------------------------------------------------------------------------
    #
    def unsubscribe(self, cb):

        for _cb, _lock in self._callbacks:
            if cb == _cb:
                self._callbacks.remove([_cb, _lock])
                break

        if not self._callbacks:
            self._stop_listener()


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._stop_listener(force=True)


    # --------------------------------------------------------------------------
    #
    def get(self):

        if not self._interactive:
            raise RuntimeError('invalid get(): callbacks are registered')


        # FIXME: add timeout to allow for graceful termination
        #
        with self._lock:
            data = no_intr(self._sock.recv)

        topic, bmsg = data.split(b' ', 1)
        msg = from_msgpack(bmsg)

        log_bulk(self._log, '<- %s' % topic, [msg])

        return [as_string(topic), as_string(msg)]


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, timeout=None):

        # FIXME:  does this duplicate _get_nowait? why / why not?

        if not self._interactive:
            raise RuntimeError('invalid get_nowait(): callbacks are registered')

        if no_intr(self._sock.poll, flags=zmq.POLLIN, timeout=timeout):

            with self._lock:
                data = no_intr(self._sock.recv, flags=zmq.NOBLOCK)

            topic, bmsg = data.split(b' ', 1)
            msg = from_msgpack(bmsg)

            log_bulk(self._log, '<- %s' % topic, [msg])

            return [as_string(topic), as_string(msg)]

        else:
            return [None, None]


# ------------------------------------------------------------------------------
# pylint: disable=unreachable
def test_pubsub(channel, addr_pub, addr_sub):

    topic = 'test'

    c_a  = 1
    c_b  = 2
    data = dict()

    for i in 'ABCD':
        data[i] = dict()
        for j in 'AB':
            data[i][j] = 0

    def cb(uid, topic, msg):
        if 'idx' not in msg:
            return
        if msg['idx'] is None:
            return False
        data[uid][msg['src']] += 1

    cb_C = lambda t,m: cb('C', t, m)
    cb_D = lambda t,m: cb('D', t, m)

    Subscriber(channel=channel, url=addr_sub, topic=topic, cb=cb_C)
    Subscriber(channel=channel, url=addr_sub, topic=topic, cb=cb_D)

    # --------------------------------------------------------------------------
    def work_pub(uid, n, delay):

        pub = Publisher(channel=channel, url=addr_pub)
        idx = 0

        while idx < n:
            time.sleep(delay)
            pub.put(topic, {'src': uid, 'idx': idx})
            idx += 1
            data[uid][uid] += 1

        # send EOF
        pub.put(topic, {'src': uid, 'idx': None})
    # --------------------------------------------------------------------------

    t_a = mt.Thread(target=work_pub, args=['A', c_a, 0.001])
    t_b = mt.Thread(target=work_pub, args=['B', c_b, 0.001])

    t_a.start()
    t_b.start()

    t_a.join()
    t_b.join()

    time.sleep(0.01)

    assert data['A']['A'] == c_a
    assert data['B']['B'] == c_b

    assert data['C']['A'] + data['C']['B'] + \
           data['D']['A'] + data['D']['B'] == 2 * (c_a + c_b)

  # print('==== %.1f %s [%s]' % (time.time(), channel, get_caller_name()))

    return data


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import regex
import collections


# ------------------------------------------------------------------------------
#
# comparison helper: convert non-iterables to list of one element
#
def _cmp_iterable(other):

    if not isinstance(other, collections.abc.Iterable):
        return [other]
    return other


# ------------------------------------------------------------------------------
#
class ReSult(object):
    '''
    This class is a container around a regular expression match, which provides
    some more conventient access methods, boolean tests, etc.

    We only handle base strings, not unicode strings!
    '''

    # -------------------------------------------------------------------------
    #
    def __init__(self, result=None):
        '''
        construct with a `regex.MatchObject` instance.  This ctor should only
        be called from within the `ReString` class.
        '''

        self._glist  = list()
        self._gdict  = dict()
        self._result = result

        if result:
            # fuck python
            if not isinstance(result, type(regex.match('',''))):
                raise TypeError('Need regex.MatchObject, not %s' % type(result))

            self._glist = result.groups()
            self._gdict = result.groupdict()


    # -------------------------------------------------------------------------
    #
    def __str__(self):
        '''
        The string representation is based on the match *list*, as the dict may
        not include all matches...
        '''

        return str(self._glist)


    # -------------------------------------------------------------------------
    #
    def __len__(self):
        '''
        The len representation is based on the match *list*, as the dict may
        not include all matches...
        '''

        return len(self._glist)


    # -------------------------------------------------------------------------
    #
    def get(self, key, default=None):
        '''
        get is supported for default based dict access,
        '''

        if isinstance(key, str):
            return self._gdict.get(key, default)
        else:
            raise TypeError('key %s needs to be integer, not %s'
                          % (key, type(key)))

    # -------------------------------------------------------------------------
    #
    def start(self, idx):

        return self._result.start(idx)


    # -------------------------------------------------------------------------
    #
    def __getitem__(self, idx):
        '''
        getitem is supported for both array type access (using an int index),
        and for dict type access (using a string name).  All other key types
        will cause an exception.
        '''

        if isinstance(idx, str):
            if idx in self._gdict:
                return self._gdict[idx]
        elif isinstance(idx, int):
            if len(self) > idx:
                return self._glist[idx]
        else:
            raise TypeError('index %s needs to be integer or string, not %s'
                           % (idx, type(idx)))
        return None


    # -------------------------------------------------------------------------
    #
    def __iter__(self):
        '''
        the matches can be iterated over
        '''

        for m in self._glist:
            yield m


    # -------------------------------------------------------------------------
    #
    def __getattr__(self, name):
        '''
        Matches can be accessed as properties
        '''

        return self[name]


    # -------------------------------------------------------------------------
    #
    def __bool__(self):
        '''
        Boolean check for 'if / elif / else' constructs
        '''

        if len(self):
            return True
        return False


    # -------------------------------------------------------------------------
    #
    def __enter__(self):
        '''
        support context manager interface for with-statement based constructs
        '''
        return self


    # -------------------------------------------------------------------------
    #
    def __exit__(self, a, b, c):
        '''
        second part of the context manager interface
        '''
        pass


    # -------------------------------------------------------------------------
    #
    # compare to another ReSult or to a tuple.  As they are both iterable, we
    # compare based on the iterable interface
    #
    #
    def __lt__(self, other):

        other = _cmp_iterable(other)
        return self < other


    # -------------------------------------------------------------------------
    #
    def __gt__(self, other):

        other = _cmp_iterable(other)
        return self > other


    # -------------------------------------------------------------------------
    #
    def __le__(self, other):

        other = _cmp_iterable(other)
        return self <= other


    # -------------------------------------------------------------------------
    #
    def __ge__(self, other):

        other = _cmp_iterable(other)
        return self >= other


    # -------------------------------------------------------------------------
    #
    def __ne__(self, other):

        other = _cmp_iterable(other)
        return not self == other


    # -------------------------------------------------------------------------
    #
    def __eq__(self, other):

        other = _cmp_iterable(other)

        if len(self) != len(other):
            return len(self) - len(other)

        for i, m in enumerate(self):
            if m != other[i]:
                print('%s != %s' % (m, other[i]))
                return m == other[i]

        return True


# ------------------------------------------------------------------------------
#
class ReString(str):
    '''
    This is a string class which supports simplified regular expression
    matching. It is not thought that the regex language or expressions are
    simplified, but rather that the invokation of the matching is simple, as is
    the handling of the match results::

        txt = ReString('The quick brown fox jumps over the lazy dog')

        # the '//' operator is overloaded to match against a regular
        # expression. The result is a `ReSult` instance, which allows simple
        # access to the matches
        with txt // r'(\\s.u)(?P<x>.*?j\\S+)' as res:
            if res: print 'Matched!'                  # boolean check
            print 'res     : '%%s' ' %% res           # list of results
            print 'res[0]  : '%%s' ' %% res[0]        # index by number ...
            print 'res[1]  : '%%s' ' %% res[1]        # ... for all matches
            print 'res['x']: '%%s' ' %% res['x']      # index by match name
            print 'res.x   : '%%s' ' %% res.x         # ...   as properties
            for i, r in enumerate(res):
                print 'res %%d   : '%%s' ' %% (i, r)  # matches as iterable

            assert len(res) == 2                     # number of matches
            assert res == [' qu', 'ick brown fox jumps']  # compare to list

        if txt // r'(rabbit)':                        # simple use in if / elif
            res = txt.get()                           # get ReSult of last match

        elif txt // r'((?:\\s).{12,15}?(\\S+))':      # full Python regex slang
            res = txt.get()

        else:
            print 'no match'
    '''

    # -------------------------------------------------------------------------
    #
    def __new__(cls, *args, **kw):

        cls._result = None
        return str.__new__(cls, *args, **kw)


    # -------------------------------------------------------------------------
    #
    def __floordiv__(self, re):

        compiled_regex = None
        if isinstance(re, str):
            compiled_regex = regex.compile(re)

        else:
            # assume we got a compiled regex
            # FIXME: type check
            compiled_regex = re

        if re:
            self._result = ReSult(compiled_regex.search(self))
            return self._result

        return None


    # -------------------------------------------------------------------------
    #
    def get(self, key=None, default=None):

        if self._result and key:

            try:
                return self._result[key]

            except KeyError:
                if default:
                    return default
                raise

        return self._result


# ------------------------------------------------------------------------------
#
def _example_re_string():

    txt = ReString('The quick brown fox jumps over the lazy dog')

    with txt // r'(\s.u)(?P<x>.*?j\S+)' as res:

        if res: print('Matched!')            # boolean check
        print('res     : %s' % res)          # list of results
        print('res[0]  : %s' % res[0])       # index by number ...
        print('res[1]  : %s' % res[1])       # ... for all matches
        print('res["x"]: %s' % res['x'])     # index by match name
        print('res.x   : %s' % res.x)        # ...   as properties
        for i, r in enumerate(res):
            print('res %d   : %s' % (i, r))  # matches as iterable

        assert len(res) == 2                # number of matches
        assert res == [' qu', 'ick brown fox jumps']  # compare to list


    if txt // '(rabbit)':                    # simple use in if / elif / ...
        res = txt.get()                      # get ReSult of last match

    elif txt // r'((?:\s).{12,15}?(\S+))':   # for full Python regex slang
        res = txt.get()

    else:
        print('no match')


# ------------------------------------------------------------------------------





import os
import sys
import time
import pprint
import signal

import threading as mt

from .misc   import as_list
from .logger import Logger


# ------------------------------------------------------------------------------
#
class Heartbeat(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, uid, timeout, interval=1, beat_cb=None, term_cb=None,
                       log=None):
        '''
        This is a simple hearteat monitor: after construction, it's `beat()`
        method needs to be called in intervals shorter than the given `timeout`
        value.  A thread will be created which checks if heartbeats arrive
        timely - if not, the current process is killed via `os.kill()` (but see
        below).

        If a callback `beat_cb` is specified, the watcher will also invoke that
        callback after every `interval` seconds.  This can be used to ensure
        heartbeats by the owning instance (which may feed this or any other
        `Heartbeat` monitor instance).  The `term_cb` callback is invoked on
        heartbeat failure, i.e., just before the process would be killed, and
        gets a single argument, the uid of the failing heartbeat sender.  If
        that term callback returns `True`, the kill is avoided though: timers
        are reset and everything continues like before.  This should be used to
        recover from failing components.

        When timeout is set to `None`,  no trigger action on missing heartbeats
        will ever be triggered.
        '''

        # we should not need to lock timestamps, in the current CPython
        # implementation, dict access is assumed to be atomic.  But heartbeats
        # should not be in the performance critical path, and should not have
        # threads competing with the (private) dict lock, so we accept the
        # overhead.

        if timeout and interval > timeout:
            raise ValueError('timeout [%.1f] too small [>%.1f]'
                            % (timeout, interval))

        self._uid      = uid
        self._log      = log
        self._timeout  = timeout
        self._interval = interval
        self._beat_cb  = beat_cb
        self._term_cb  = term_cb
        self._term     = mt.Event()
        self._lock     = mt.Lock()
        self._tstamps  = dict()
        self._pid      = os.getpid()
        self._watcher  = None

        if not self._log:
            self._log  = Logger('radical.utils.heartbeat')

        self._log.debug_1('hb %s create', self._uid)


    # --------------------------------------------------------------------------
    #
    def start(self):

        self._log.debug_1('hb %s start', self._uid)
        self._watcher  = mt.Thread(target=self._watch)
        self._watcher.daemon = True
        self._watcher.start()


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._term.set()

      # # no need to join, is a daemon thread
      # self._watcher.join()


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        return self._uid


    # --------------------------------------------------------------------------
    #
    def dump(self, log):

        if not log: log = self._log

        log.debug_9('hb dump %s: \n%s', self._uid, pprint.pformat(self._tstamps))


    # --------------------------------------------------------------------------
    #
    def watch(self, uid):

        with self._lock:
            if uid not in self._tstamps:
                self._log.debug_8('hb %s watch %s', self._uid, uid)
                self._tstamps[uid] = None


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        # initial heartbeat without delay
        if  self._beat_cb:
            self._log.debug_8('hb %s beat cb init', self._uid)
            self._beat_cb()

        while not self._term.is_set():

            self._log.debug_8('hb %s loop %s', self._uid, self._interval)

            time.sleep(self._interval)
            now = time.time()

            if  self._beat_cb:
                self._log.debug_8('hb %s beat cb', self._uid)
                self._beat_cb()

            # avoid iteration over changing dict
            with self._lock:
                uids = list(self._tstamps.keys())

            self._log.debug_8('hb %s uids %s', self._uid, uids)
            for uid in uids:

                self._log.debug_8('hb %s check %s', self._uid, uid)

                with self._lock:
                    last = self._tstamps.get(uid)

                if last is None:
                    self._log.warn('hb %s inval %s', self._uid, uid)
                    continue

                if now - last > self._timeout:

                    if self._log:
                        self._log.warn('hb %s tout  %s: %.1f - %.1f > %.1f',
                                       self._uid, uid, now, last, self._timeout)

                    ret = None
                    if self._timeout:
                        # attempt to recover
                        if  self._term_cb:
                            ret = self._term_cb(uid)
                    else:
                        # we silently assume that the watchee recovered, thus
                        # avoiding termination
                        ret = True

                    if ret in [None, False]:
                        # could not recover: abandon mothership
                        self._log.warn('hb %s fail  %s: fatal (%d)',
                                       self._uid, uid, self._pid)
                        os.kill(self._pid, signal.SIGTERM)
                        time.sleep(0.1)
                        os.kill(self._pid, signal.SIGKILL)

                    else:
                        # recovered - the failed UID was replaced with the one
                        # returned by the callback.  We delete the heartbeat
                        # information for the old uid and register a new
                        # heartbeat for the new one, so that we can immediately
                        # begin to watch it.
                        assert isinstance(ret, str)
                        self._log.info_1('hb %s recov %s -> %s (%s)',
                                          self._uid, uid, ret, self._term_cb)
                        with self._lock:
                            del self._tstamps[uid]
                            self._tstamps[ret] = time.time()


    # --------------------------------------------------------------------------
    #
    def beat(self, uid=None, timestamp=None):

        if not timestamp:
            timestamp = time.time()

        if not uid:
            uid = 'default'

        with self._lock:
            self._log.debug_9('hb %s beat [%s]', self._uid, uid)
            self._tstamps[uid] = timestamp


  # # --------------------------------------------------------------------------
  # #
  # def is_alive(self, uid=None):
  #     '''
  #     Check if an entity of the given UID sent a recent heartbeat
  #     '''
  #
  #     if not uid:
  #         uid = 'default'
  #
  #     with self._lock:
  #         ts = self._tstamps.get(uid)
  #
  #     if ts and time.time() - ts <= self._timeout:
  #         return True
  #
  #     return False


    # --------------------------------------------------------------------------
    #
    def wait_startup(self, uids=None, timeout=None):
        '''
        Wait for the first heartbeat of the given UIDs to appear.  This returns
        the list of UIDs which have *not* been found, or `None` otherwise.
        '''

        if not uids:
            uids = ['default']

        uids = as_list(uids)

        start = time.time()
        ok    = list()
        while True:

            with self._lock:
                ok  = [uid for uid in uids if self._tstamps.get(uid)]
                nok = [uid for uid in uids if uid not in ok]

            self._log.debug_3('wait for : %s', nok)

            if len(ok) == len(uids):
                break

            if timeout:
                if time.time() - start > timeout:
                    self._log.debug_3('wait time: %s', nok)
                    break

            time.sleep(0.25)

        if len(ok) != len(uids):
            nok = [uid for uid in uids if uid not in ok]
            self._log.error('wait fail: %s', nok)
            return nok

        else:
            self._log.debug_3('wait ok  : %s', ok)


# ------------------------------------------------------------------------------
#
class PWatcher(object):

    NOTHING = 'nothing'
    SUICIDE = 'suicide'
    KILLALL = 'killall'
    RAMPAGE = 'rampage'

    # --------------------------------------------------------------------------
    #
    def __init__(self, action=None, uid=None, log=None):
        '''

        This is a simple process monitor: once started it runs in a separate
        thread and monitors all given process IDs (`self._watch_pid`).  If
        a process is found to have died, the watcher will invoke the given
        action:

          - `nothing`: log event and do nothing else
          - `suicide`: kill the curent process
          - `killall`: kill all monitored pids
          - `rampage`: both of the above (`suicide + killall`)

        The default action is `rampage`.

        The passed uid (default: `pwatcher`) is used for logging purposes only.
        '''

        self._action = action or self.RAMPAGE
        self._uid    = uid    or 'pwatcher'
        self._log    = log    or Logger(name=self._uid, ns='radical.utils')
        self._pids   = list()
        self._lock   = mt.Lock()

        self._log.debug_1('pwatcher create')

        self._thread = mt.Thread(target=self._watch)
        self._thread.daemon = True
        self._thread.start()


    # --------------------------------------------------------------------------
    #
    def _is_alive(self, pid):

        try           : os.kill(pid, 0)
        except OSError: return False
        else          : return True


    # --------------------------------------------------------------------------
    #
    def _send_signal(self, pid, sig, group=False):

        if self._is_alive(pid):
            try:
                if group: os.killpg(pid, sig)
                else    : os.kill(pid, sig)
            except OSError as e:
                pass

            # let signal handler kick in
            time.sleep(0.05)


    # --------------------------------------------------------------------------
    #
    def _kill(self, pid):

        try   : os.killpg(pid, signal.SIGTERM)
        except: pass

        time.sleep(0.05)
        try   : os.kill(pid, signal.SIGTERM)
        except: pass

        # pytest will barf at any SIGKILL
        if os.environ.get("PYTEST_VERSION") is None:

            time.sleep(0.05)
            try   : os.killpg(pid, signal.SIGKILL)
            except: pass

            time.sleep(0.05)
            try   : os.kill(pid, signal.SIGKILL)
            except: pass


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        self._log.debug_1('pwatcher started')

        while True:

            with self._lock:

                for pid in list(self._pids):

                    if not self._is_alive(pid):

                        self._log.warn('process %d died, exit', pid)
                        self._pids.remove(pid)

                        if   self._action == self.NOTHING: self._nothing(pid)
                        elif self._action == self.SUICIDE: self._suicide(pid)
                        elif self._action == self.KILLALL: self._killall(pid)
                        elif self._action == self.RAMPAGE: self._rampage(pid)

            time.sleep(0.05)


    # --------------------------------------------------------------------------
    #
    def watch(self, pid):

        self._log.debug('add pid %d to watchlist', pid)

        with self._lock:
            self._pids.append(pid)


    # --------------------------------------------------------------------------
    #
    def unwatch(self, pid):

        self._log.debug('remove pid %d from watchlist', pid)

        with self._lock:
            if pid in self._pids:
                self._pids.remove(pid)

    # --------------------------------------------------------------------------
    #
    def _nothing(self, pid):

        self._log.debug("process %d's demise triggered, well, nothing", pid)


    # --------------------------------------------------------------------------
    #
    def _suicide(self, pid):

        self._log.debug("process %d's demise triggered suicide", pid)
        self._kill(os.getpid())


    # --------------------------------------------------------------------------
    #
    def _killall(self, pid):

        self._log.debug("process %d's demise triggered killall (%s)",
                        pid, self._pids)

        for pid in list(self._pids):
            self._kill(pid)


    # --------------------------------------------------------------------------
    #
    def _rampage(self, pid):

        self._killall(pid)
        self._suicide(pid)


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import time
import threading

from .lockable import Lockable
from .logger   import Logger

# default settings for lease manager
MAX_POOL_SIZE = 15       # unlimited
MAX_POOL_WAIT = 60       # seconds
MAX_OBJ_AGE   = 60 * 10  # 10 minutes


# ------------------------------------------------------------------------------
#
class _LeaseObject(object):

    _uid = 0

    # --------------------------------------------------------------------------
    def __init__(self, lm, log, creator, args):

        self.lm         = lm
        self.used       = False
        self.log        = log
        self.obj        = creator(*args)
        self.uid        = 'lo.%04d' % _LeaseObject._uid
        self.t_created  = time.time()
        self.t_leased   = None
        self.t_released = time.time()  # we take control *now*

        _LeaseObject._uid += 1


    # --------------------------------------------------------------------------
    #
    def __cmp__(self, other):

        return bool(other == self.obj)


    # --------------------------------------------------------------------------
    #
    def __enter__(self):

        return self.obj


    # --------------------------------------------------------------------------
    #
    def __exit__(self, *args):                                      # noqa W0613

        self.lm.release(self)


    # --------------------------------------------------------------------------
    #
    def lease(self):

        if self.used:
            raise RuntimeError('LeaseObject is already leased: %s' % self)

        self.used     = True
        self.t_leased = time.time()


    # --------------------------------------------------------------------------
    #
    def release(self):

        if not self.used:
            raise RuntimeError('LeaseObject is not leased: %s' % self)

        self.used       = False
        self.t_released = time.time()


    # --------------------------------------------------------------------------
    def is_leased(self):

        return self.used


# ------------------------------------------------------------------------------
#
@Lockable
class LeaseManager(object):
    '''
    This is a lease manager -- it creates resource instances on demand and hands
    out leases to them.  If for a given ID no object instance exists, one is
    created, locked and a lease is returned.  If for an ID an object instance
    exists and is not leased, it is locked and returned.  If one or more
    instances exist but all are in use already (leased), a new instance is
    created (up to MAX_POOL_SIZE which can be overwritten in the lease call).
    If that limit is reached, no objects are returned, and instead the lease
    call blocks until one of the existing objects gets released.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, max_pool_size=None, max_pool_wait=None,
                       max_obj_age=None):
        '''
        Make sure the object dict is initialized, exactly once.
        '''

        self._log = Logger('radical.utils')

        self._log.debug('lm new manager')
        self._pools = dict()

        self._max_pool_size = max_pool_size
        self._max_pool_wait = max_pool_wait
        self._max_obj_age   = max_obj_age

        if self._max_pool_size is None: self._max_pool_size = MAX_POOL_SIZE
        if self._max_pool_wait is None: self._max_pool_wait = MAX_POOL_WAIT
        if self._max_obj_age   is None: self._max_obj_age   = MAX_OBJ_AGE


    # --------------------------------------------------------------------------
    #
    def _initialize_pool(self, pool_id):
        '''
        set up a new pool, but do not create any instances, yet.
        '''

        with self:

            if pool_id not in self._pools:

                self._log.debug('lm create  pool   for %s (%s) (%s)'
                        % (pool_id, type(pool_id), self))

                self._pools[pool_id] = dict()
                self._pools[pool_id]['objects'] = list()
                self._pools[pool_id]['freed']   = None
                self._pools[pool_id]['event']   = threading.Event()

            return self._pools[pool_id]


    # --------------------------------------------------------------------------
    #
    def _create_object(self, pool_id, creator, args):
        '''
        a new instance is needed -- create one, unless max_pool_size is reached.
        If that is the case, return `None`, otherwise return the created object
        (which is locked before return).
        '''

        with self:

            self._log.debug('lm create  object for %s' % pool_id)

            if  pool_id not in self._pools:
                raise RuntimeError('internal error: no pool for %s' % pool_id)

            pool = self._pools[pool_id]

            # check if a poolsize cap is set and reched
            if  self._max_pool_size >  0 and \
                self._max_pool_size <= len(pool['objects']):

                # no more space...
                return None

            # poolsize cap not reached -- increase pool.  If creating a new
            # object does not work for any reason, return None.
            obj = None
            try:
                obj = _LeaseObject(self, self._log, creator, args)
                obj.lease()
                pool['objects'].append(obj)

            except Exception as e:
                # this exception needs to fall through -- we can't wait
                # for object creation problems to fix themself over time...
                self._log.exception('Could not create lease object: %s', e)
                raise

            return obj


    # --------------------------------------------------------------------------
    #
    def _remove_object(self, pool_id, obj):
        '''
        a new instance is needed -- create one, unless max_pool_size is reached.
        If that is the case, return `None`, otherwise return the created object
        (which is locked before return).
        '''

        with self:

            if  pool_id not in self._pools:
                raise RuntimeError('internal error: no pool for %s' % pool_id)

            pool = self._pools[pool_id]

            # poolsize cap not reached -- increase pool.  If creating a new
            # object does not work for any reason, return None.
            obj = None
            try:
                pool['objects'].remove(obj)

            except Exception as e:
                self._log.warn('Could not remove lease object: %s', e)

            return obj


    # --------------------------------------------------------------------------
    #
    def lease(self, pool_id, creator, args=None):
        '''
        For a given object identified, attempt to retrieve an existing object
        from the pool.  If such a free (released) object is found, lock and
        refturn it.  If such objects exist but all are in use, create a new one
        up to max_pool_size (default: 10).
        used, block untill it is freed.  If that object does not exist, create
        it and proceed per above.
        return the object thusly created.

        pool_id       : id of the pool to lease from.  The pool ID essentially
                        determines the scope of validity for the managed objects
                        -- they form a namespace, where objects are shared when
                        living under the same name space entry (the pool_id).

        creator       : method to use to create a new object instance

                        Example:
                            def creator() :
                                return Logger(name)

                            ret = lease_manager.lease(name, creator)
        '''

        pool_id = str(pool_id)

        if   not args                  : args = []
        elif not isinstance(args, list): args = [args]

        with self:

            # make sure the pool exists
            pool = self._initialize_pool(pool_id)

            # find an unlocked object instance in the pool
            # NOTE: we iterate over a copy of the list, as an eventual object
            # removeal would screw up the index...
            for obj in pool['objects'][:]:

                if  not obj.is_leased():

                    # check age
                    age = time.time() - obj.t_created
                    if  age > self._max_obj_age:
                        # too old -- remove and continue to search for
                        # a younger unleased object
                        self._remove_object(pool_id, obj)
                        continue

                    # found one -- lease/lock and return it
                    obj.lease()
                    return obj

            # no unlocked object found -- create a new one
            obj = self._create_object(pool_id, creator, args)

            # FIXME: we could try_catch the above error, and then check if the
            #        pool has anything worth to wait on.  That might be useful
            #        for creation errors which are transient.  Alas, we don't
            #        have any means to distinguish them from non-transient
            #        errors, so we don't do that at this point...

            # check if we got an object
            if  obj is not None:

                # we got a locked object -- return
                return obj


        # pool is full, nothing is free -- we need to wait for an event on
        # the pool to free an instance.
        # Now, this is where deadlocks will happen: any application leasing
        # too many instances in the same thread (with 'too many' meaning
        # more than max_pool_size) will lock that thread here, thus having no
        # chance to release other instances.  We thus will print a log error
        # here, and will raise a timeout exception after MAX_POOL_WAIT
        # seconds.
        # Not that we release our lock here, to give other threads the
        # chance to release objects.
        self._log.warning('lm lease   object: pool is full')
        timer_start = time.time()
        timer_now   = time.time()

        while(timer_now - timer_start) < self._max_pool_wait:

            # wait for any release activity on the pool
            timer_left = self._max_pool_wait - (timer_now - timer_start)
            pool['event'].wait(timer_left)

            with self:

                if  pool['event'].is_set():

                    # make sure we don't lock up
                    pool['event'].clear()

                    # we won the race!  now get the freed object
                    obj = pool['freed']
                    pool['freed'] = None

                    if  obj is None:

                        # object was deleted -- we should have space to create
                        # a new one!
                        obj = self._create_object(pool_id, creator, args)

                        # check if we got an object
                        if  obj is not None:

                            # we got a locked object -- return
                            return obj

                        else:

                            # find did not find a freed object, did not get to
                            # create a new one -- is there a free one in the
                            # pool?
                            for obj in pool['objects']:

                                if  not obj.is_leased():

                                    # found one -- lease/lock and return it
                                    obj.lease()
                                    return obj

                            # apparently not - so we give up for now...


                    else:
                        # we got a freed object -- lock and return
                        obj.lease()
                        return obj


                # none free, none created - or we lost the rase for handling the
                # event.  Wait again.
                timer_now = time.time()


        # at this point we give up: we can't create a new object, can't find
        # a free one, and we are running out of wait time...
        raise LookupError('stop waiting on object lease')


    # --------------------------------------------------------------------------
    #
    def release(self, instance, delete=False):
        '''
        the given object is not needed right now -- unlock it so that somebody
        else can lease it.  This will not delete the object,
        '''

        with self:

            for pool_id in self._pools:

                for obj in self._pools [pool_id]['objects']:
                    if  instance is not obj:
                        # this is not the object you are looking for.
                        continue

                    # remove the lease lock on the object
                    obj.release()

                    if  delete:
                        # remove the object from the pool (decreasing its
                        # ref counter and thus making it eligible for
                        # garbage collection).
                        self._pools [pool_id]['objects'].remove(obj)
                        self._pools [pool_id]['freed'] = None

                    else:
                        # mark the object as freed for lease.
                        self._pools [pool_id]['freed'] = obj

                    # notify waiting threads about the lease or creation
                    # opportunity.
                    self._pools [pool_id]['event'].set()

                    # object has been released
                    return

            # FIXME: log warning
            # for now we ignore double-frees
            # raise RuntimeError('cannot release object -- not managed')


# ------------------------------------------------------------------------------






import os
import time
import errno

import threading as mt

from .misc   import as_bytes, ru_open
from .debug  import get_caller_name


# ------------------------------------------------------------------------------
#
class Lockfile(object):
    '''
    This class represents a lockfile.  On calling `open()` (or entering
    a resource context `with`), the lockfile is exclusively opened and locked.
    The returned object can be used for `read()`, `write()` and `seek()`
    operations, and the lock is only released on `close()` (or when leaving the
    resource context).

    If `delete=True` is specified on construction, then the lockfile is removed
    uppon `release()`, and the content is lost.  Only for this mode is it
    possible to obtain the current lock owner when using `get_owner()` while
    waiting for a lock owned by another process or thread - the call will return
    'unknown' otherwise.

    Note that the file offset is not shared between owners of the lock file
    - a newly acquired lock on the file should prompt a seek to the desired
    file location (the initial offset is '0').

    Example::

        with ru.Lockfile(fname) as lf0:
            lf0.write('test 0\\n')

        with ru.Lockfile(fname) as lf1:
            lf1.lseek(0, os.SEEK_SET)
            lf1.write('test 1\\n')

        with ru.Lockfile(fname) as lf2:
            lf2.lseek(0, os.SEEK_END)
            lf2.write('test 2\\n')

            # raises a RuntimeError, as we won't be able to acquire the lock
            with ru.Lockfile(fname) as lf3:
                lf3.lseek(0, os.SEEK_END)
                lf3.write('test 3\\n')

        with open(fname, 'r') as fin:
            data = fin.read()
            assert data == 'test 1\\ntest 2\\n'

    Example::

        lf1 = ru.Lockfile(fname)
        lf2 = ru.Lockfile(fname)

        with lf1:
            lf1.write('test 0')

            with lf2(timeout=3):
                # we won't reach this, the `with` above will time out and raise
                lf2.write('test 0')


    Implementation:

    The implementation relies on cooperative locking: a process or thread which
    uses this API to lock a file will create a temporary file in the same
    directory (filename: `.<filename>.<pid>.<counter>` where `counter` is
    a process singleton).  On lock acquisition, we will attempt to create
    a symbolic link to the file `<filename>.lck`, which will only succeed if no
    other process or thread holds that lock.  On `release`, that symbolic gets
    removed; on object destruction, the temporary file gets removed.

    The private file will contain the name of the owner.  Any other process or
    thread can follow the `.<filename>.lck` symlink, open the link's target
    file, and read the name of the owner.  No attempt is made to avoid races
    between failure to acquire the lockfile and querying the current owner, so
    that information is not reliable, but intented to be an informative help for
    debugging and tracing purposes.

    Example::
        lf1 = ru.Lockfile(fname)
        lf2 = ru.Lockfile(fname)

        with lf1:
            lf1.write('test 0')

            try:
                lf2.acquire()  # timeout == None == 0.0
            except RuntimeError:
                print('lock is held by %s' % lf2.get_owner())
    '''

    _counter = 0
    _lock    = mt.Lock()


    # --------------------------------------------------------------------------
    #
    def __init__(self, fname, *args, **kwargs):
        '''
        The `args` and `kwargs` arguments are passed to `acquire()` when used in
        a `with Lockfile():` clause:

            with ru.Lockfile(fname, timeout=3) as lockfile:
                lockfile.write(data)

        '''

        kwargs.setdefault('delete', False)

        self._fname   = fname
        self._fd      = None

        if 'delete' in kwargs:
            self._delete = kwargs['delete']
            del kwargs['delete']
        else:
            self._delete = False

        self._args    = args
        self._kwargs  = kwargs

        # create a tempfile to be used for lock acquisition
        with Lockfile._lock:
            cnt = Lockfile._counter
            Lockfile._counter += 1


        fname_base = os.path.basename(self._fname)
        fname_dir  = os.path.dirname(self._fname)

        self._lck  = '%s/%s.lck'    % (fname_dir, fname_base)
        self._tmp  = '%s/.%s.%d.%d' % (fname_dir, fname_base, os.getpid(), cnt)

        # make sure our tmp file exists
        with ru_open(self._tmp, 'w') as fout:
            fout.write('%s\n' % get_caller_name())


    # --------------------------------------------------------------------------
    #
    def __call__(self, *args, **kwargs):
        '''
        helper method to pass arguments while using the `with lock` clause:

            with lock(timeout=2, owner=foo):
                lock.write(data)
        '''

        self._args   = args
        self._kwargs = kwargs

        return self


    # --------------------------------------------------------------------------
    #
    def __enter__(self):

        self.acquire(*self._args, **self._kwargs)
        return self


    # --------------------------------------------------------------------------
    #
    def __exit__(self, foo, bar, baz):

        return self.release()


    # --------------------------------------------------------------------------
    #
    def __bool__(self) -> bool:

        return bool(self._fd)


    # --------------------------------------------------------------------------
    #
    def locked(self) -> bool:

        return bool(self)


    # --------------------------------------------------------------------------
    #
    def acquire(self, timeout=None, owner=None):
        '''
        When the lock could not be acquired after `timeout` seconds, an
        `TimeoutError` is raised (i.e. an `OSError` with errno `ETIME`).  When
        `owner` is specified and the lockfile was created with `delete=True`,
        then that string will be passed to any other method which tries to
        acquire this lockfile while it is locked (see `get_owner()`).  If not
        specified, the `owner` string is set to `ru.get_caller_name()`.

        timeout:
            < 0 : wait forever
            = 0 : try locking once and raise `RuntimeError` on failure
            > 0 : try for that many seconds, and raise `TimeoutError` on failure
            None: same as 0

        '''

        if timeout is None:
            timeout = 0.0

        if self._fd:
            raise RuntimeError('cannot call open twice')

        # pre-emptively record who wants to acquire the lock
        if not owner:
            owner = get_caller_name()

        with ru_open(self._tmp, 'w') as fout:
            fout.write('%s\n' % owner)

        start = time.time()
        while True:
            # attempt to link self._tmp to self._lck.  Once that succeeds, open
            # the self._fnmame for read/write.

            try:
                os.symlink(self._tmp, self._lck)
                break  # link succeeded, we have the lock

            # if self._lck exists the above `link` will raise, and we try
            # again after a bit.  All other errors are fatal
            except OSError as e:

                # pylint: disable=W0707
                if not e.errno == errno.EEXIST:
                    raise

                if timeout == 0.0:
                    raise RuntimeError('failed to lock %s (%s)'
                                       % (self._fname, self.get_owner()))

                elif timeout > 0:
                    now = time.time()
                    if now - start > timeout:
                        raise TimeoutError(errno.ETIME, 'lock timeout for %s ()'
                                           % self._fname, self.get_owner())
                # try again
                time.sleep(0.1)  # FIXME: granulatiy?
                continue

        # the link succeeded, so open the file and break
        self._fd = os.open(self._fname, os.O_RDWR | os.O_CREAT)

        return self.locked()


    # --------------------------------------------------------------------------
    #
    def get_owner(self):
        '''
        That name of the current owner of the lockfile can be retrieved by any
        other process or thread which attempts to lock the same file.  This
        method returns `None` if the file is not currently locked.

        Note: this method may race against lock acquisition / release, and the
              information returned may be outdated.
        '''

        try:
            with ru_open(self._lck, 'r') as fin:
                # strip newline
                return fin.readline()[:-1]

        except OSError as e:
            if e.errno == errno.EEXIST:
                return None
            else:
                return 'unknown'


    # --------------------------------------------------------------------------
    #
    def release(self):
        '''
        Release the lock on the file.  If `delete=True` was specified on
        construction, then the file (and all owner information) are removed.
        Once released, all other threads/processes waiting in the `acquire()`
        method call will compete for the lock, and one of them will obtain it.
        '''

        if not self._fd:
            raise RuntimeError('lockfile is not open')

        # release the file handle
        os.close(self._fd)
        self._fd = None

        if self._delete:
            self.remove()

        # release the lock
        os.unlink(self._lck)


    # --------------------------------------------------------------------------
    #
    def remove(self):

        os.unlink(self._fname)


    # --------------------------------------------------------------------------
    #
    def read(self, length):
        '''
        Read from the locked file at the current offset.  This method will raise
        an `RuntimeError` when being called without the file being locked.
        '''

        if not self._fd:
            raise RuntimeError('lockfile is not open')

        return os.read(self._fd, length)


    # --------------------------------------------------------------------------
    #
    def write(self, data):
        '''
        Write to the locked file at the current offset.  This method will raise
        an `RuntimeError` when being called without the file being locked.
        '''

        if not self._fd:
            raise RuntimeError('lockfile is not open')

        return os.write(self._fd, as_bytes(data))


    # --------------------------------------------------------------------------
    #
    def seek(self, pos, how):
        '''
        Same as `lseek()`
        '''

        return self.lseek(pos, how)


    # --------------------------------------------------------------------------
    #
    def lseek(self, pos, how):
        '''
        Change the offset at which the next read or write method is applied.
        The arguments are interpreted as documented by `os.lseek()`.
        '''

        if not self._fd:
            raise RuntimeError('lockfile is not open')

        return os.lseek(self._fd, pos, how)


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky, Ole Weidner)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import os
import sys


from .modules import import_module
from .json_io import read_json


# ------------------------------------------------------------------------------
#
# when running under `pytest`, `sys.exit()` will trigger a test error, even if
# it is an expected call.  Mocking or monkeypatching `sys.exit()` removes the
# error, but will alter semantics so that the test becomes meaningless.  We thus
# switch to `pytest.exit()` if running under pytest.
#
# It is not easy to determine if we are running under pytest.  We thus assume
# that we are the only instance checking for that, and that the `pytest` module
# is imported somewhere in the interpreter before `ru` is imported.
_pytest_active = False

# check if pytest is loaded
for m in sys.modules:
    if 'pytest' in str(m):
        _pytest_active = True
        break

# check if we can import pytest
if _pytest_active:
    try:
        import pytest
    except ImportError:
        _pytest_active = False


def sys_exit(ret):
    '''
    call `pytest.exit(ret)` when running under pytest, `sys.exit(ret)` otherwise
    '''

    if _pytest_active:
        pytest.exit(reason='Exit with pytest', returncode=ret)

    else:
        sys.exit(ret)

    assert False


# ------------------------------------------------------------------------------
#
# The currently active test configuration
#
_test_config = dict()


# ------------------------------------------------------------------------------
#
def get_test_config():
    '''
    If a test config is currently set, return it.  If not, attempt to load atest
    config for the given namespace, set is as active, and return it.
    '''

    return _test_config


# ------------------------------------------------------------------------------
#
def set_test_config(ns, cfg_name=None, cfg_section=None):
    '''
    Set a test config.  All subsequent calls to `get_test_config()` will
    retrieve the same configuration, until a new config is explicitly set with
    this method again.
    '''

    global _test_config                                  # pylint: disable=W0603
    _test_config = TestConfig(ns, cfg_name, cfg_section)


# ------------------------------------------------------------------------------
#
def add_test_config(ns, cfg_name, cfg_section=None):
    '''
    To an existing active config, add the contents of an additional test
    configuration.
    '''

    if not _test_config:
        raise RuntimeError('cannot add config, no active config is set')

    _test_config.add_config(ns, cfg_name, cfg_section)


# ------------------------------------------------------------------------------
#
class TestConfig(dict):
    '''
    This class represents a set of configurations for a test suite.  It usually
    contains parameters which tweak the tests toward specific environments and
    test cases, such as specific remote systems to interact with, specific
    backends to use, etc.

    The class expects test configurations to be deployed in the current Python
    installation (virtualenv, condaenv, etc.).  The namespace passed on
    construction is used to search for configurations under
    `$namespace/configs/tests/`, configuration files are expected to follow the
    naming scheme `test_$cfg_name.json`.  The json file is read and this object
    then behaves like a dictionary for those json data, but all top level keys
    in the dictionary are also exposed as object attributes.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, ns, cfg_name=None, cfg_section=None):

        if not cfg_name:
            cfg_name = 'default'

        self._ns  = ns
        self._cfg = self._load_config(ns, cfg_name, cfg_section)

        dict.__init__(self, self._cfg)

        self._initialized = True  # see self.__setattr__()


    # --------------------------------------------------------------------------
    #
    def _load_config(self, ns, cfg_name, cfg_section):

        mod   = import_module(ns)
        path  = '%s/configs/tests/' % os.path.dirname(mod.__file__)
        fname = '%s/test_%s.json'   % (path, cfg_name)

        if not os.path.isfile(fname):
            raise ValueError('no such config file %s' % fname)

        cfg = read_json(fname)

        if cfg_section:

            if cfg_section not in cfg:
                raise ValueError('no such config section %s' % cfg_section)

            cfg = cfg[cfg_section]

        return cfg


    # --------------------------------------------------------------------------
    #
    def add_config(self, ns, cfg_name, cfg_section):
        '''
        To the current config dict contents, add the content of the specified
        additional config file
        '''

        if ns != self._ns:
            raise ValueError('cannot merge configs from different name spaces')

        cfg = self._load_config(ns, cfg_name, cfg_section)

        for k,v in cfg.items():
            self._cfg[k] = v

        dict.__init__(self, self._cfg)


    # --------------------------------------------------------------------------
    #
    def __getattr__(self, item):
        '''
        default to None instead of raising
        '''
        try:
            return self.__getitem__(item)

        except KeyError:
            return None


    # --------------------------------------------------------------------------
    #
    def __setattr__(self, item, value):
        '''
        Maps attributes to values
        '''

        # allow attributes to be set in the __init__ method
        if self._initialized:
            return dict.__setattr__(self, item, value)

        # known attributes
        elif item in self.__dict__:
            dict.__setattr__(self, item, value)

        # default
        else:
            self.__setitem__(item, value)


# ------------------------------------------------------------------------------




# pylint: disable=reimported

# allow star import, unused symbols
# flake8: noqa: F401

__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2013-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

# ------------------------------------------------------------------------------
#
import os            as _os


# ------------------------------------------------------------------------------
#
# we want atfork imported first, specifically before os and logging
#
from .atfork         import *

# import constants
from .constants      import *

# import utility classes
from .object_cache   import ObjectCache
from .plugin_manager import PluginManager, PluginBase
from .singleton      import Singleton
from .heartbeat      import Heartbeat, PWatcher
from .threads        import is_main_thread, is_this_thread, cancel_main_thread
from .threads        import main_thread, this_thread, get_thread_name, gettid
from .threads        import set_cancellation_handler, unset_cancellation_handler
from .threads        import raise_in_thread, ThreadExit, SignalRaised
from .futures        import Future
from .futures        import NEW, RUNNING, DONE, FAILED, CANCELED
from .url            import Url
from .host           import get_hostip, get_hostname
from .host           import get_hostlist, get_hostlist_by_range
from .host           import create_hostfile, compress_hostlist, is_localhost
from .lockable       import Lockable
from .lockfile       import Lockfile
from .registry       import Registry, READONLY, READWRITE
from .ru_regex       import ReString, ReSult
from .lease_manager  import LeaseManager
from .daemon         import Daemon, daemonize
from .poll           import Poller, POLLIN, POLLOUT, POLLERR, POLLALL
from .poll           import POLLNVAL, POLLPRI, POLLHUP
from .shell          import sh_quote
from .shell          import sh_callout, sh_callout_bg, sh_callout_async
from .testing        import sys_exit
from .testing        import TestConfig
from .testing        import set_test_config, add_test_config, get_test_config
from .env            import env_read, env_write, env_read_lines, env_eval
from .env            import env_prep, env_diff, EnvProcess, env_dump
from .stack          import stack
from .modules        import import_module, find_module, import_file
from .modules        import get_type, load_class

from .dict_mixin     import DictMixin, dict_merge, dict_stringexpand, dict_diff
from .dict_mixin     import PRESERVE, OVERWRITE, iter_diff
from .typeddict      import TypedDict, TypedDictMeta, as_dict
from .config         import Config, DefaultConfig

from .zmq            import Message
from .zmq            import Bridge
from .zmq            import Queue,  Putter,    Getter
from .zmq            import PubSub, Publisher, Subscriber
from .zmq            import Server, Client

from .flux           import FluxService, FluxHelper

from .logger         import DEBUG, INFO, WARNING, WARN, ERROR, CRITICAL, OFF
from .logger         import Logger
from .reporter       import Reporter
from .profile        import Profiler, timestamp, event_to_label
from .profile        import read_profiles, combine_profiles, clean_profile
from .profile        import TIME, EVENT, COMP, TID, UID, STATE, MSG, ENTITY
from .profile        import PROF_KEY_MAX
from .profile        import Yappi

from .json_io        import read_json, read_json_str, write_json
from .json_io        import parse_json, parse_json_str, dumps_json
from .which          import which
from .tracer         import trace, untrace
from .get_version    import get_version

from .serialize      import to_json, from_json, to_msgpack, from_msgpack
from .serialize      import register_serializable


# import various utility methods
from .ids            import *
from .debug          import *
from .misc           import *
from .algorithms     import *

# import decorators
from .timing         import timed_method, epoch, dt_epoch, Time

# import sub-modules
from . import scheduler
from . import config
from . import zmq


# ------------------------------------------------------------------------------
#
# get version info
#
_mod_root = _os.path.dirname (__file__)

version_short, version_base, version_branch, version_tag, version_detail \
             = get_version(_mod_root)
version      = version_short
__version__  = version_detail


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
import zmq
import time
import msgpack

DELAY = 0.0


# ------------------------------------------------------------------------------
#
addr = None
with open('test.bridge.url', 'r') as fin:
    for line in fin.readlines():
        tag, addr = line.split()
        if tag == 'GET':
            break

print('GET: %s' % addr)

context    = zmq.Context()
socket     = context.socket(zmq.REQ)
socket.hwm = 1
socket.connect(addr)

while True:
    socket.send('request %d' % os.getpid())
    msg = msgpack.unpackb(socket.recv())
    print('<- %s' % msg)
    time.sleep(DELAY)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"

import os
import time
import select
import threading as mt


# `select.poll()` is a nice thing - alas, it is not implemented in MacOS'
# default python deployment.  Since RU process management relies on socket
# polling, we reimplement it here based on `select.select()`.
#
# The following import directive should enable any select based code to run
# unchanged on MacOS (for the parts we port, at least):
#
#     import radical.utils.poll as select
#
#     poller = select.poll()
#     ...
#
# If `RU_USE_PYPOLL` is set (to an arbitrary, non-empty value) in the
# environment, we fall back to the native Python implementation.
#
_use_pypoll = os.environ.get('RU_USE_PYPOLL', False)


# ------------------------------------------------------------------------------
# define the Poller factory.  No idea why the Poller object is not directly
# exposed in the `select` module... :/
def poll(log=None):
    return (Poller(log))


# ------------------------------------------------------------------------------
if _use_pypoll:

    POLLIN   = select.POLLIN
    POLLPRI  = select.POLLPRI
    POLLOUT  = select.POLLOUT
    POLLERR  = select.POLLERR
    POLLHUP  = select.POLLHUP
    POLLNVAL = select.POLLNVAL

    POLLALL  = POLLIN | POLLOUT | POLLERR | POLLPRI | POLLHUP | POLLNVAL

    # --------------------------------------------------------------------------
    #
    class Poller(object):
        '''
        This is a this wrapper around select.Poller to have it API compatible
        with our own, select based implementation.
        '''

        def __init__(self, log=None):

            self._log  = log
            self._poll = select.poll()


        def close(self):
            self._poll = None


        def register(self, fd, eventmask=None):
            assert self._poll
            return self._poll.register(fd, eventmask)


        def modify(self, fd, eventmask=None):
            assert self._poll
            return self._poll.modify(fd, eventmask)


        def unregister(self, fd):
            assert self._poll
            return self._poll.unregister(fd)


        def poll(self, timeout=None):
            assert self._poll
            time.sleep(0.2)
            ret = self._poll.poll(timeout)
          # if self._log:
          #     self._log.debug('pypoll: %s', ret)
            return ret


# ------------------------------------------------------------------------------
#
else:
    # these defines are magically compatible with the ones in Python's `select`
    # module...
    POLLIN   = 0b000001
    POLLPRI  = 0b000010
    POLLOUT  = 0b000100
    POLLERR  = 0b001000
    POLLHUP  = 0b010000
    POLLNVAL = 0b100000

    # POLLALL is not defined by `select`.
    POLLALL  = POLLIN | POLLOUT | POLLERR | POLLPRI | POLLHUP | POLLNVAL

    # we only really support the following 4 types - all others are silently
    # ignored, and will never be triggered.
    _POLLTYPES = [POLLIN, POLLOUT, POLLERR, POLLHUP]


    # --------------------------------------------------------------------------
    #
    class Poller(object):
        '''
        This object will accept a set of things we can call `select.select()`
        on, which is basically anything which has a file desciptor exposed via
        a `fileno()` member method, or any integers which directly represent an
        OS level file escriptor.  This class implements most of the interface
        as defined by Python's `select.Poller` class, which is created by
        calling `select.poll()`.  This implementation is thread-safe.

        NOTE: `Poller.poll()` returns the original object handle instead of the
              polled file descriptor.
        NOTE: Support for `POLLPRI` and `POLLNVAL` selection is not implemented,
              and will never be returned on `poll()`.
        '''

        # ----------------------------------------------------------------------
        #
        def __init__(self, log=None):

            self._log        = log
            self._lock       = mt.RLock()
            self._registered = {POLLIN  : list(),
                                POLLOUT : list(),
                                POLLERR : list(),
                                POLLHUP : list()}


        # ----------------------------------------------------------------------
        #
        def register(self, fd, eventmask=None):

            if not eventmask:
                eventmask = POLLIN | POLLOUT | POLLERR

            with self._lock:

                self.unregister(fd, _assert_existence=False)

                for e in _POLLTYPES:
                    if eventmask & e:
                        self._registered[e].append(fd)


        # ----------------------------------------------------------------------
        #
        def close(self):

            with self._lock:

                for e in _POLLTYPES:
                    for fd in self._registered[e]:
                        fd.close()
                    self._registered[e] = list()


        # ----------------------------------------------------------------------
        #
        def _exists(self, fd):

            with self._lock:
                for e in _POLLTYPES:
                    if fd in self._registered[e]:
                        return True

            return False


        # ----------------------------------------------------------------------
        #
        def modify(self, fd, eventmask=None):

            if not eventmask:
                eventmask = POLLIN | POLLOUT | POLLERR | POLLHUP

            with self._lock:
                assert self._exists(fd)
                self.register(fd, eventmask)


        # ----------------------------------------------------------------------
        #
        def unregister(self, fd, _assert_existence=True):

            with self._lock:

                exists = self._exists(fd)

                if _assert_existence:
                    assert exists
                elif not exists:
                    return

                for e in _POLLTYPES:
                    if fd in self._registered[e]:
                        self._registered[e].remove(fd)


        # ----------------------------------------------------------------------
        #
        def poll(self, timeout=None):

            ret = list()

            with self._lock:
                rlist = self._registered[POLLIN]
                wlist = self._registered[POLLOUT]
                xlist = self._registered[POLLERR]
                hlist = self._registered[POLLHUP]

                # only select if we have any FDs to watch
                if not rlist + wlist + xlist + hlist:
                    return ([], [], [])

                # Des Pudel's Kern!
                rret, wret, xret = select.select(rlist + hlist, wlist,
                                                 xlist, timeout)

              # if self._log:
              #     self._log.debug('rppoll 0 : %s - %s - %s', rret, wret, xret)

                # do not return hlist-only FDs
                ret += [[fd, POLLOUT] for fd in set(wret)]
                ret += [[fd, POLLERR] for fd in set(xret)]

                # A socket being readable may also mean the emdpoint has been
                # closed.  We do a zero-length read to check for POLLHUP.
                #
                # NOTE: I am so happy we can do type inspection in Python, so
                #       that we can have different checks for, say, files and
                #       sockets.  Only this shit doesn't work on sockets! Argh!
                #       Oh well, we now guess the type from `recv` and `read`
                #       methods being available.
                #       Its a shame to do this on every `poll()` call, as this
                #       is likely in a performance critical path...
                for fd in set(rret):

                    hup = False

                    # file object
                    if hasattr(fd, 'closed'):
                        if fd.closed:
                            hup = True

                    # anything with a `fileno()` and `read()`/`write()`
                    elif hasattr(fd, 'read'):
                        try:
                            fd.read(0)
                            fd.write(b'')
                        except Exception:
                            hup = True

                    # socket
                    elif hasattr(fd, 'recv'):
                        try:
                            fd.recv(0)
                            fd.send(b'')
                        except Exception:
                            hup = True

                    # we can't handle errors on other types
                    else:
                        raise TypeError('cannot check %s [%s]' % (fd, type(fd)))

                    # we always return the POLLIN event, too, as that is what
                    # Python natively does.  I don't think it makes sense,
                    # semantically, but whatever...
                    if hup: ret.append([fd, POLLIN | POLLHUP])
                    else  : ret.append([fd, POLLIN])

              # if self._log:
              #     self._log.debug('rppoll 1 : %s (%s)', ret, rlist)

                return ret


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'

import sys
import os

from functools import reduce


# ------------------------------------------------------------------------------
# This file (ab)uses a number of python features which are disliked by pylint
# pylint: disable=E1120,W1505,W0613

# ------------------------------------------------------------------------------
#
# Method call parameters/return value type checking decorators.
# (c) 2006-2007, Dmitry Dvoinikov <dmitry@targeted.org>
# Distributed under BSD license.
#
# code adjustements by 'The SAGA Project', 2013
# See also
# http://code.activestate.com/recipes/\
#        577065-type-checking-function-overloading-decorator/
#
#
# Samples:
#
# from typecheck import *
#
# @takes(int, str) # takes   int, str, upon a problem throws TypeError
# @returns(int)    # returns int,      upon a problem throws TypeError
# def foo(i, s):
#     return i + len(s)
#
# @takes((int, long), by_regex('^[0-9]+$')) # int or long, numerical string
# def foo(i, s, anything):                  # third parameter is not checked
#     ...
#
# @takes(int, int, foo = int, bar = optional(int)) # keyword foo must be int
# def foo(a, b, **kwargs):                         # bar may be int or missing
#     ...
#
# Note: @takes for positional arguments, @takes for keyword arguments and
# @returns all support the same checker syntax, for example for the following
# declaration
#
# @takes(C)
# def foo(x):
#     ...
#
# then C may be one of the simple checkers:
#
# --------- C ---------     ------------- semantics -------------
# typename              ==> ok if x is is an instance of typename
# 'typename'            ==> ok if x is is an instance of typename
# with_attr('a', 'b')   ==> ok if x has specific attributes
# some_callable         ==> ok if some_callable(x) is True
# one_of(1, '2')        ==> ok if x is one of the literal values
# by_regex('^foo$')     ==> ok if x is a matching basestring
# nothing               ==> ok if x is None
# anything              ==> always ok
#
# simple checkers can further be combined with OR semantics using tuples:
#
# --------- C ---------     ------------- semantics -------------
# (checker1, checker2)      ==> ok if x conforms with either checker
#
# be optional:
#
# --------- C ---------     ------------- semantics -------------
# optional (checker)        ==> ok if x is checker-conformant or None
#
# or nested recursively into one of the following checkers
#
# --------- C ---------     ------------- semantics -------------
# list_of  (checker)        ==> ok if x is a list  of checker-conformant values
# tuple_of (checker)        ==> ok if x is a tuple of checker-conformant values
# set_of   (checker)        ==> ok if x is a set   of checker-conformant values
# dict_of  (key_checker,
#           value_checker)  ==> ok if x is a dict  mapping key_checker-
#                           conformant keys to value_checker-conformant values
#
# More samples:
#
# class foo(object):
#     @takes('foo', optional(int)) # foo, maybe int, but foo is yet incomplete
#     def __init__(self, i = None): # and is thus specified by name
#         ...
#     @takes('foo', int)            # foo, and int if presents in args,
#     def bar(self, *args):         # if args is empty, the check passes ok
#         ...
#     @takes('foo')
#     @returns(object)              # returns foo which is fine, because
#     def biz(self):                # foo is an object
#         return self
#     @classmethod                   # classmethod's and staticmethod's
#     @takes(type)                  # go same way
#     def baz(cls):
#         ...
#
# @takes(int)
# @returns(optional('int', foo))   # returns either int, foo or NoneType
# def bar(i):                       # 'int' (rather than just int) is for fun
#     if i > 0:
#         return i
#     elif i == 0:
#         return foo()               # otherwise returns NoneType
#
# @takes(callable)                # built-in funcs are treated as predicates
# @returns(lambda x: x == 123)      # and so do user-defined funcs or lambdas
# def exec(f, *args, **kwargs):
#     return f(*args, **kwargs)
#
# assert execute(execute, execute, execute, lambda x: x, 123) == 123
#
# def readable(x):                  # user-defined type-checking predicate
#     return hasattr(x, 'read')
#
# anything is an alias for predicate lambda: True,
# nothing is an alias for NoneType, as in:
#
# @takes(callable, readable,
#           optional(anything),
#           optional(int))
# @returns(nothing)
# def foo(f, r, x = None, i = None):
#     ...
#
#                                    # another way of protocol checking
# @takes(with_attr('read', 'write'))
# def foo(pipe):
#     ...
#
# @takes(list_of(int))             # list of ints
# def foo(x):
#     print x[0]
#
# @takes(tuple_of(callable))       # tuple of callables
# def foo(x):
#     print x[0]()
#
#                                    # dict mapping strs to lists of int
# @takes(dict_of(str, list_of(int)))
# def foo(x):
#     print sum(x['foo'])
#
# @takes(by_regex('^[0-9]{1,8}$')) # integer-as-a-string regex
# def foo(x):
#     i = int(x)
#
# @takes(one_of(1, 2))             # must be equal to either one
# def set_version(version):
#     ...
#
# The(3 times longer) source code with self-tests is available from:
# http://www.targeted.org/python/recipes/typecheck.py
#
# ------------------------------------------------------------------------------

__all__ = ['takes',    'returns',   'optional', 'nothing',
           'anything', 'list_of',   'tuple_of', 'dict_of',
           'by_regex', 'with_attr', 'one_of',   'set_of' ]

no_check        = True  # set this to True to turn all checks off
no_return_check = True   # set this to True to turn return value cchecks off

if 'RADICAL_DEBUG_SIG' in os.environ:
    no_check = False  # set this to True to turn all checks off

# ------------------------------------------------------------------------------

from traceback import extract_stack
from re        import compile as regex
from functools import wraps
from inspect   import getfullargspec, isclass

# NOTE: Python 3: type NoneType does not exist anymore. Test against None.
# from types     import NoneType
NoneType = type(None)


# ------------------------------------------------------------------------------
#
def base_names(C):
    '''
    Returns list of base class names for a given class
    '''

    return [x.__name__ for x in C.__mro__]


# ------------------------------------------------------------------------------
#
def type_name(v):
    '''
    Returns the name of the passed value's type
    '''

    return type(v).__name__


# ------------------------------------------------------------------------------
#
class Checker(object):

    registered = list()  # a list of registered descendant class factories

    def __init__(self, reference):
        self.reference = reference
        self.spectype  = reference


    def check(self, value):  # abstract
        pass


    @staticmethod
    def create(value):  # static factory method

        for f, t in Checker.registered:
            if f(value):
                return t(value)


# ------------------------------------------------------------------------------
#
class TypeChecker(Checker):

    def check(self, value):
        return isinstance(value, self.reference)


Checker.registered.append((isclass, TypeChecker))
nothing = NoneType


# ------------------------------------------------------------------------------
#
class StrChecker(Checker):

    def check(self, value):
        value_base_names = base_names(type(value))
        return self.reference in value_base_names or \
               'instance'     in value_base_names


Checker.registered.append((lambda x: isinstance(x, str), StrChecker))


# ------------------------------------------------------------------------------
#
class TupleChecker(Checker):

    def __init__(self, reference):
        self.reference = list(map(Checker.create, reference))
        self.spectype  = reference

    def check(self, value):
        return reduce(lambda r, c: r or c.check(value), self.reference, False)


Checker.registered.append((lambda x: isinstance(x, tuple) and not
                     [y for y in x if Checker.create(y) is None],
                     TupleChecker))


def optional(*args):
    return args + (NoneType, )


# ------------------------------------------------------------------------------
#
class CallableChecker(Checker):

    def check(self, value):
        return self.reference(value)

# note that the callable check is the most relaxed of all, therefore it should
# be registered last, after all the more specific cases have been registered


Checker.registered.append((callable, CallableChecker))


def anything(*args):
    return True


# ------------------------------------------------------------------------------
#
class ListOfChecker(Checker):

    def __init__(self, reference):
        self.reference = Checker.create(reference)

    def check(self, value):
        return isinstance(value, list) and \
               not [e for e in value if not self.reference.check(e)]


def list_of(*args):
    return ListOfChecker(*args).check


# ------------------------------------------------------------------------------
#
class TupleOfChecker(Checker):

    def __init__(self, reference):
        self.reference = Checker.create(reference)

    def check(self, value):
        return isinstance(value, tuple) and \
               not [e for e in value if not self.reference.check(e)]


def tuple_of(*args):
    return TupleOfChecker(*args).check


# ------------------------------------------------------------------------------
#
class SetOfChecker(Checker):

    def __init__(self, reference):
        self.reference = Checker.create(reference)

    def check(self, value):
        return isinstance(value, set) and \
               not [e for e in value if not self.reference.check(e)]


def set_of(*args):
    return SetOfChecker(*args).check


# ------------------------------------------------------------------------------
#
class DictOfChecker(Checker):

    def __init__(self, key_reference, value_reference):
        self.key_reference   = Checker.create(key_reference)
        self.value_reference = Checker.create(value_reference)

    def check(self, value):
        return isinstance(value, dict) and \
               not [e for e in value.keys()
                            if not self.key_reference.check(e)] and \
               not [e for e in value.values()
                            if not self.value_reference.check(e)]


def dict_of(*args):
    return DictOfChecker(*args).check


# ------------------------------------------------------------------------------
#
class RegexChecker(Checker):

    def __init__(self, reference):
        self.reference = regex(reference)

    def check(self, value):
        return isinstance(value, str) and self.reference.match(value)


def by_regex(*args):
    return RegexChecker(*args).check


# ------------------------------------------------------------------------------
#
class AttrChecker(Checker):

    def __init__(self, *attrs):
        self.attrs = attrs

    def check(self, value):
        return reduce(lambda r, c: r and c,
                  [hasattr(value, a) for a in self.attrs], True)


def  with_attr(*args):
    return AttrChecker(*args).check


# ------------------------------------------------------------------------------
#
class OneOfChecker(Checker):

    def __init__(self, *values):
        self.values = values

    def check(self, value):
        return value in self.values


def one_of(*args):
    return OneOfChecker(*args).check


# ------------------------------------------------------------------------------
#
def create_return_exception(method, spectype, result):

    stack = extract_stack()
    for f in stack:
        if  'utils/signatures.py' in f[0]:
            break
        frame = f

    msg   = '\nSignature Mismatch\n'
    msg  += '  in function   : %s\n'     % (frame[2])
    msg  += '  in file       : %s +%s\n' % (frame[0], frame[1])
    msg  += '  on line       : %s\n'     % (frame[3])
    msg  += '  method        : %s\n'     % (method.__name__)
    msg  += '  returned type : %s\n'     % (type(result).__name__)
    msg  += '  instead  of   : %s\n'     % (spectype.__name__)
    msg  += '  This is an internal error!'

    return TypeError(msg)


# ------------------------------------------------------------------------------
#
def create_type_exception(method, arg0, i, arg, spectype, kwname=''):

    narg = 0

    if  arg0 and isinstance(arg0, object):
        narg = i
    else:
        narg = i + 1

    stack = extract_stack()
    for f in stack:
        if  'utils/signatures.py' in f[0]:
            break
        frame = f

    if isinstance(spectype,tuple) or \
       isinstance(spectype,list ):
        expected = list()
        for t in spectype:
            if isinstance(t, type):
                expected.append(t.__name__)
            else:
                expected.append(str(t))

    elif isinstance(spectype,type):
        expected = spectype.__name__

    else:
        expected = str(spectype)

    msg   = '\nSignature Mismatch\n'
    msg  += '  in function        : %s\n'       % (frame[2])
    msg  += '  in file            : %s +%s\n'   % (frame[0], frame[1])
    msg  += '  on line            : %s\n'       % (frame[3])
    msg  += '  method             : %s\n'       % (method.__name__)
    if  not kwname :
        msg  += '  argument           : #%s\n'  % (narg)
    else :
        msg  += '  parameter          : %s'     % (kwname)
    msg  += '  has incorrect type : %s\n'       % (type(arg).__name__)
    msg  += '  instead of         : %s'         % (str(expected))

    return TypeError(msg)


# ------------------------------------------------------------------------------
#
def takes(*args, **kwargs):
    '''
    Method signature checking decorator
    '''

    if no_check:  # no type check is performed, return decorated method itself

        def takes_proxy(method):
            return method

    else:

        # convert decorator arguments into a list of checkers
        checkers = list()
        for i, arg in enumerate(args):
            checker = Checker.create(arg)
            if checker is None:
                raise TypeError('@takes got parameter %d of unsupported '
                                 'type %s' % (i + 1, type_name(arg)))
            checkers.append(checker)

        kwcheckers = dict()
        for kwname, kwarg in kwargs.items():
            checker = Checker.create(kwarg)
            if checker is None:
                raise TypeError('@takes got parameter %s of unsupported '
                                 'type %s' % (kwname, type_name(kwarg)))
            kwcheckers[kwname] = checker


        def takes_proxy(method):

            method_args, method_defaults = getfullargspec(method)[0::3]

            @wraps(method)
            def signature_check(*pargs, **pkwargs):

                # append the default parameters
                if  method_defaults is not None and len(method_defaults) > 0 \
                    and len(method_args) - len(method_defaults) <= \
                                                  len(pargs) < len(method_args):
                    pargs += method_defaults[len(pargs) - len(method_args):]

                # check the types of the actual call parameters
                for i,(arg, checker) in enumerate(zip(pargs, checkers)):

                    if  not checker.check(arg):
                      # print 'Checker.spectype %s' % checker.spectype
                        excpt = create_type_exception(method, pargs[0], i,
                                                      arg, checker.spectype)
                        raise excpt

                for kwname, checker in kwcheckers.items():

                    if  not checker.check(pkwargs.get(kwname, None)):
                      # print 'checker.spectype %s' % checker.spectype
                        # pylint: disable=W0631
                        excpt = create_type_exception(method, pargs[0], i,
                                                      arg, checker.spectype,
                                                      kwname)
                        raise excpt

                try:
                    return method(*pargs, **pkwargs)

                except Exception as e:
                    # remove signature decorator from exception call stack
                    et, ei, tb = sys.exc_info()
                    raise et(ei).with_traceback(tb.tb_next) from e

            signature_check.__name__ = method.__name__
            return signature_check

    return takes_proxy


# ------------------------------------------------------------------------------
#
def returns(sometype):
    '''
    Return type checking decorator
    '''

    if no_check:

        def returns_proxy(method):
            return method

    else:

        # convert decorator argument into a checker
        checker = Checker.create(sometype)
        if checker is None:
            raise TypeError('@returns decorator got parameter of unsupported '
                             'type %s' % type_name(sometype))

        def returns_proxy(method):

            @wraps(method)
            def signature_check(*args, **kwargs):

                try:
                    result = method(*args, **kwargs)
                except Exception as e:
                    # remove signature decorator from exception call stack
                    et, ei, tb = sys.exc_info()
                    raise et(ei).with_traceback(tb.tb_next) from e

                if  not checker.check(result):

                    if not  no_return_check:
                      # print 'Checker.spectype %s' % checker.spectype
                        excpt = create_return_exception(method,
                                                       checker.spectype, result)
                        raise excpt

                return result

            signature_check.__name__ = method.__name__
            return signature_check

    return returns_proxy


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

import re
import os
import sys
import queue
import hashlib
import tempfile
import traceback

from typing import List, Dict, Tuple, Any, Optional

import multiprocessing as mp

from .misc  import as_list, rec_makedir, ru_open
from .shell import sh_callout


# we know that some env vars are not worth preserving.  We explicitly exclude
# those which are common to have complex syntax and need serious caution on
# shell escaping:
IGNORE_LIST = ['PS1', 'LS_COLORS', '_', 'SHLVL', 'PROMPT_COMMAND', 'LINES',
               'COLUMNS']

# Identical task `pre_exec_cached` settings will result in the same environment
# settings, so we cache those environments here.  We rely on a hash to ensure
# `pre_exec_cached` identity.  Note that this assumes that settings do not
# depend on, say, the unit ID or similar, which needs very clear and prominent
# documentation.  Caching can be turned off by adding a unique noop string to
# the `pre_exec_cached` list - but we probably also add a config flag if that
# becomes a common issue.
_env_cache = dict()

# we use a regex to match snake_case words which we allow for variable names
# with the following conditions
#   - starts with a letter or underscore
#   - consists of letters, underscores and numbers
re_snake_case = re.compile(r'^[a-zA-Z_]\w*$', re.ASCII)

# regex to check if a variable refers to a bash shell function
re_bash_function = re.compile(r'^BASH_FUNC_([a-zA-Z_]\w+)(%%|\(\))$', re.ASCII)

# regex to detect the start of a function definition as written by `env_write()`
re_function = re.compile(r'^([a-zA-Z_]\w+)\(\) {$', re.ASCII)


# ------------------------------------------------------------------------------
#
def env_read(fname: str) -> Dict[str, str]:
    '''
    helper to parse environment from a file: this method parses the output of
    `env` and returns a dict with the found environment settings.
    '''

    with ru_open(fname, 'r') as fin:
        lines = fin.readlines()

    return env_read_lines(lines)


# ------------------------------------------------------------------------------
#
def env_write(script_path, env, unset=None, blacklist=None, pre_exec=None,
              extend=False):

    data = '\n'
    if unset:
        data += '# unset\n'
        for k in sorted(unset):
            if k in env:
                continue
            if not re_snake_case.match(k):
                continue
            data += 'unset %s\n' % k
        data += '\n'

    if IGNORE_LIST:
        data += '# blacklist\n'
        for k in sorted(IGNORE_LIST):
            data += 'unset %s\n' % k
        data += '\n'

        env = {k: v for k, v in env.items() if k not in IGNORE_LIST}

    if blacklist:
        data += '# blacklist\n'
        for k in sorted(blacklist):
            data += '# %s\n' % k
        data += '\n'

        # apply blacklist
        bl_exact   = [k for k in blacklist if not k.endswith('*')]
        env = {k: v for k, v in env.items() if k not in bl_exact}

        bl_pattern = [k for k in blacklist if k.endswith('*')]
        for p in bl_pattern:
            env = {k: v for k, v in env.items() if not k.startswith(p[:-1])}

    funcs = list()
    data += '# export\n'
    for k in sorted(env.keys()):
        if k.startswith('BASH_FUNC_') and \
                (k.endswith('%%') or k.endswith('()')):
            funcs.append(k)
            continue
        if not re_snake_case.match(k):
            continue
        data += "export %s=%s\n" % (k, _quote(env[k], extend=extend))
    data += '\n'

    if funcs:
        data += '\n# functions\n'
        for func in funcs:
            fname = func[10:-2]
            v     = env[func]
            if v.startswith('() { '):
                v = v.replace('() { ', '() {\n', 1)
            data += '%s%s\n' % (fname, v.replace('\\n', '\n'))
            data += 'test -z "$BASH" || export -f %s\n\n' % fname
        data += '\n'

    if pre_exec:
        data += '# pre_exec (not cached)\n'

        # do not sort, order dependent
        for cmd in pre_exec:
            data += '%s\n' % cmd
        data += '\n'

    with ru_open(script_path, 'w') as fout:
        fout.write(data)


# ------------------------------------------------------------------------------
#
def env_read_lines(lines: List[str]) -> Dict[str, str]:
    '''
    read lines which are the result of an `env` shell call, and sort the
    resulting keys into and environment and a shell function dict, return both
    '''

    # POSIX definition of variable names
    env     = dict()
    key     = None
    val     = ''

    for line in lines:

        # remove newline
        line = line.rstrip('\n')

        if not line:
            continue

        # search for new key
        if '=' not in line:
            # no key present - append linebreak and line to value
            val += '\n'
            val += line
            continue

        elems    = line.split('=', 1)
        this_key = elems.pop(0)
        this_val = elems[0] if elems else ''

        if re_snake_case.match(this_key):
            # valid key - store previous key/val if we have any, and
            # initialize `key` and `val`
            if key and key not in IGNORE_LIST:
                env[key] = val

            key = this_key
            val = this_val

        elif re_bash_function.match(this_key):
            # function definitions
            # initialize `key` and `val`
            if key and key not in IGNORE_LIST:
                env[key] = val

            key = this_key
            val = this_val

        else:
            # invalid key - append linebreak and line to value
            val += '\n'
            val += line

    # store last key/val if we have any
    if key and key not in IGNORE_LIST:
        env[key] = val

    return env


# ------------------------------------------------------------------------------
#
def _quote(data: str, extend: bool = False) -> str:

    if "'" in data or '$' in data or '`' in data:
        # cannot use single quote, so use double quote and escale all other
        # double quotes in the data
        # NOTE: we only support these three types of shell directives
        data = data.replace('"', '\\"')
        if not extend:
            data = data.replace('$', '\\$')
        data = '"' + data + '"'

    else:
        # single quotes will do
        data = "'" + data +  "'"

    return data


# ------------------------------------------------------------------------------
#
def _unquote(data: str) -> str:

    if data.startswith("'") and data.endswith("'"):
        # just remove enclosing single quotes - no nesting
        data = data[1:-1]

    elif data.startswith('"') and data.endswith('"'):
        # remove enclosing double quotes, and replace all occurences of escaled
        # double quotes (`\"`) with an unescaled one (`"`).
        data = data[1:-1]
        data = data.replace('\\"', '"')

    return data


# ------------------------------------------------------------------------------
#
def env_eval(fname: str) -> Dict[str, str]:
    '''
    helper to create a dictionary with the env settings in the specified file
    which contains `unset` and `export` directives, or simple 'key=val' lines
    '''

    env = dict()
    with ru_open(fname, 'r') as fin:

        func_name = None
        func_data = None
        for line in fin.readlines():

            if func_name:
                # we capture a function definition right now - check if done
                if line == 'test -z "$BASH" || export -f %s' % func_name:
                    # done - convert into a bash env variables
                    env['BASH_FUNC_%s%%%%' % func_name] = '() %s' % func_data
                    # stop function parsing
                    func_name = None
                    func_data = None
                    continue
                else:
                    # still part of function data, replace the stripped newline
                    if func_data:
                        func_data += '\n'
                    func_data += line
                    continue

            line = line.strip()

            if not line:
                continue

            if line.startswith('#'):
                continue

            func_check = re_function.match(line)
            if func_check:
                # detected start of function
                assert func_name is None
                func_name = func_check[1]
                func_data = ''
                assert func_name
                continue


            if line.startswith('unset ') :
                _, spec = line.split(' ', 1)
                k = spec.strip()
                if k not in env:
                    continue
                del env[k]

            elif line.startswith('export ') :
                _, spec = line.split(' ', 1)
                elems   = spec.split('=', 1)
                k = elems.pop(0)
                v = elems[0] if elems else ''
                env[k] = _unquote(v.strip())

            else:
                elems = line.split('=', 1)
                k = elems.pop(0)
                v = elems[0] if elems else ''
                env[k] = _unquote(v.strip())

    return env


# ------------------------------------------------------------------------------
#
def env_dump(environment: Optional[Dict[str,str]] = None,
             script_path: Optional[str]           = None) -> None:

    if not environment:
        environment = dict(os.environ)

    if script_path:
        with ru_open(script_path, 'w') as fout:
            for k in sorted(environment.keys()):
                fout.write('%s=%s\n' % (k, environment[k].replace('\n', '\\n')))
    else:
        for k in sorted(environment.keys()):
            print('%s=%s' % (k, environment[k].replace('\n', '\\n')))


# ------------------------------------------------------------------------------
#
def env_prep(environment    : Optional[Dict[str,str]] = None,
             unset          : Optional[List[str]]     = None,
             blacklist      : Optional[List[str]]     = None,
             pre_exec       : Optional[List[str]]     = None,
             pre_exec_cached: Optional[List[str]]     = None,
             script_path    : Optional[str]           = None
            ) -> Dict[str, str]:
    '''
    Create a shell script which restores the environment specified in
    `environment` environment (dict).  While doing so, ensure that all env
    variables *not* defined in `environment` but defined in `unset` (list) are
    unset.  Also ensure that all commands provided in `pre_exec_cached` (list)
    are executed after these settings.

    If `blacklist` is given, all variables listed there will not to be touched,
    so will neither be unset nor set.  This is useful to ensure that certain
    values are not changed, even if they are not part of the `environment` dict.
    `blacklist` can also contain patterns like `FOO_*` to blacklist all
    variables starting with `FOO_`.

    Once the shell script is created, run it and dump the resulting env, then
    read it back via `env_read()` and return the resulting env dict - that
    can then be used for process fork/execve to run a process is the thus
    defined environment.

    The resulting environment will be cached: a subsequent call with the same
    set of parameters will simply return a previously cached environment if it
    exists.

    If `script_path` is given, a shell script will be created in the given
    location so that shell commands can source it and restore the specified
    environment.

    Any commands given in 'pre_exec' will be part of the cached script, and will
    thus *not* be executed when preparing the env, but *will* be executed
    whenever the prepared shell script is sources.  The returned env dictionary
    will thus *not* include the effects of those injected commands.
    '''

    # defaults
    if environment     is None: environment     = dict(os.environ)
    if unset           is None: unset           = list()
    if blacklist       is None: blacklist       = list()
    if pre_exec        is None: pre_exec        = list()
    if pre_exec_cached is None: pre_exec_cached = list()

    if pre_exec and not script_path:
        raise ValueError('`pre_exec` must be used with `script_path`')

    # empty `pre_exec*` settings are ok - just ensure correct type
    pre_exec        = as_list(pre_exec       )
    pre_exec_cached = as_list(pre_exec_cached)

    # cache lookup
    cache_key = str(sorted(environment.items())) \
              + str(sorted(unset))               \
              + str(sorted(pre_exec_cached))
    cache_md5 = hashlib.md5(cache_key.encode('utf-8')).hexdigest()

    if cache_md5 in _env_cache:
        env = _env_cache[cache_md5]

    else:
        # cache miss

        # Write a temporary shell script which
        #
        #   - unsets all variables which are not defined in `environment`
        #     but are defined in the `unset` list
        #   - unset all `unset` vars
        #   - sets all variables defined in the `environment` dict
        #   - runs the `pre_exec_cached` commands given
        #   - dumps the resulting env in a temporary file
        #
        # Then run that script and read the resulting env back into a dict to
        # return.  If `script_path` is specified, then also create a file at the
        # given name and fill it with `unset` and `export` statements to
        # recreate that specific environment: any shell sourcing that
        # `script_path` file thus activates the environment we just prepared.
        tgt = os.getcwd() + '/env/'
        rec_makedir(tgt)

        if script_path: prefix = os.path.basename(script_path)
        else          : prefix = None

        _, tmp_name = tempfile.mkstemp(prefix=prefix, dir=tgt)

        env_write(tmp_name, env=environment, unset=unset, blacklist=blacklist,
                            pre_exec=pre_exec_cached, extend=False)
        cmd = '/bin/bash -c ". %s && /usr/bin/env"' % tmp_name
        out, err, ret = sh_callout(cmd)

        if ret:
            raise RuntimeError('error running "%s": %s' % (cmd, err))

        env = env_read_lines(out.split('\n'))
        os.unlink(tmp_name)

        # cache the resulting env
        _env_cache[cache_md5] = env


    # If `script_path` is specified, create a script with that name which unsets
    # the same names as in the tmp script above, and exports all vars from the
    # resulting env from above (thus storing the *results* of the
    # `pre_exec_cached` env, not the env and `pre_exec_cached` directives
    # themselves).
    #
    # FIXME: files could also be cached and re-used (copied or linked)
    if script_path:
        env_write(script_path, env=env, unset=unset, blacklist=blacklist,
                  pre_exec=pre_exec, extend=True)

    return env


# ------------------------------------------------------------------------------
#
def env_diff(env_1 : Dict[str,str],
             env_2 : Dict[str,str]
            ) -> Tuple[Dict[str,str], Dict[str,str], Dict[str,str]]:
    '''
    This method serves debug purposes: it compares to environments and returns
    those elements which appear in only either one or the other env, and which
    changed from one env to another.

    It will ignore any keys in the `IGNORE_LIST` and will also ignore
    `BASH_FUNC_*` keys which point to bash function definitions.
    '''

    only_1  = dict()
    only_2  = dict()
    changed = dict()

    keys_1 = sorted(env_1.keys())
    keys_2 = sorted(env_2.keys())

    for k in keys_1:
        if k in IGNORE_LIST:
            continue
        if k.startswith('BASH_FUNC_'):
            continue
        v = env_1[k]
        if   k not in env_2: only_1[k]  = v
        elif v != env_2[k] : changed[k] = [v, env_2[k]]

    for k in keys_2:
        if k in IGNORE_LIST:
            continue
        if k.startswith('BASH_FUNC_'):
            continue
        v = env_2[k]
        if k not in env_1: only_2[k]  = v
        # else is checked in keys_1 loop above

    return only_1, only_2, changed


# ------------------------------------------------------------------------------
#
class EnvProcess(object):
    '''
    run a code segment in a different os.environ::

        env = {'foo': 'buz'}
        with ru.EnvProcess(env=env) as p:
            if p:
                p.put(os.environ['foo'])

        print('-->', p.get())
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, env : Dict[str, str]) -> None:

        self._q     = mp.Queue()
        self._env   = env
        self._data  = None
        self._child = None


    # --------------------------------------------------------------------------
    #
    def __bool__(self) -> Optional[bool]:

        return self._child


    # --------------------------------------------------------------------------
    #
    def __enter__(self) -> 'EnvProcess':

        if os.fork():
            self._parent = True
            self._child  = False
        else:
            self._parent = False
            self._child  = True

        if self._child:

            for k in os.environ:
                del os.environ[k]

            for k, v in self._env.items():
                os.environ[k] = v

            # refresh the python interpreter in that environment
            import site
            import importlib

            importlib.reload(site)
            importlib.invalidate_caches()

        return self


    # --------------------------------------------------------------------------
    #
    def __exit__(self, exc_type: Optional[Exception],
                       exc_val : Optional[Any],
                       exc_tb  : Optional[Any]
                ) -> None:

        if self._parent:
            while self._data is None:
                try               : self._data = self._q.get(timeout=1)
                except queue.Empty: pass

        elif exc_type and self._child:
            stacktrace = ' '.join(traceback.format_exception(
                                                     exc_type, exc_val, exc_tb))
            self._q.put([None, exc_type, exc_val, stacktrace])
            self._q.close()
            self._q.join_thread()
            os._exit(0)


    # --------------------------------------------------------------------------
    #
    def put(self, data: str) -> None:

        assert self._child

        self._q.put([data, None, None, None])
        self._q.close()
        self._q.join_thread()
        os._exit(0)


    # --------------------------------------------------------------------------
    #
    def get(self) -> Any:

        assert self._parent

        if self._data is None:
            return

        data, exc_type, exc_val, stacktrace = self._data
        if exc_type:
            sys.stderr.write('envp excepted %s(%s)\n' % (exc_type, exc_val))
            sys.stderr.write(stacktrace)
            sys.stderr.flush()
            raise RuntimeError('envp failed: %s(%s) - check stderr'
                               % (exc_type, exc_val))

        return data


# ------------------------------------------------------------------------------





import os
import sys
import glob

from .misc    import as_list
from .shell   import sh_callout
from .modules import import_module


# ------------------------------------------------------------------------------
#
def stack(ns='radical'):
    '''
    returns a dict with information about the currently active python
    interpreter and all radical modules (incl. version details)
    '''

    exe = sh_callout('which python3', shell=True)[0].strip()
    ret = {'sys'     : {'python'     : exe,
                        'version'    : sys.version.split()[0],
                        'pythonpath' : os.environ.get('PYTHONPATH',  ''),
                        'virtualenv' : os.environ.get('VIRTUAL_ENV', '') or
                                       os.environ.get('CONDA_DEFAULT_ENV','')},
          }

    for namespace in as_list(ns):

        ret[namespace] = dict()

        ns_mod = import_module(namespace)
        path   = ns_mod.__path__
        if isinstance(path, list):
            path = path[0]

        if isinstance(path, str):
            rpath = path
        else:
            rpath = path._path                               # pylint: disable=W0212

        if isinstance(rpath, list):
            rpath = rpath[0]

        for mpath in glob.glob('%s/*' % rpath):

            if os.path.isdir(mpath):

                mbase = os.path.basename(mpath)
                mname = '%s.%s' % (namespace, mbase)

                if mbase.startswith('_'):
                    continue

                try:
                    mod = import_module(mname)
                    version        = getattr(mod, 'version')
                    version_detail = getattr(mod, 'version_detail')

                    if version_detail:
                        ret[namespace][mname] = version_detail
                    else:
                        ret[namespace][mname] = version

                except Exception as e:
                    ret[namespace][mname] = 'Error: %s' % str(e)

    return ret


# ------------------------------------------------------------------------------





# flake8: noqa: F401

__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2018"
__license__   = "GPL"


from .bridge   import Bridge
from .queue    import Queue,  Putter,    Getter,     test_queue
from .pubsub   import PubSub, Publisher, Subscriber, test_pubsub
from .pipe     import Pipe,   MODE_PUSH, MODE_PULL
from .client   import Client
from .server   import Server
from .registry import Registry, RegistryClient
from .message  import Message


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
os.environ['RADICAL_DEBUG'] = 'TRUE'

import time
import radical.utils as ru


# ------------------------------------------------------------------------------
#
# This example demonstrates the DebugHelper class.  Once started, use the
# printed pid to send a SIGUSR1 like this:
#
#   kill -USR1 <pid>
#
# and the DebugHelper will print stack traces for all threads to stdout.
#
# ------------------------------------------------------------------------------
#
def worker_outer():
    worker_inner()


def worker_inner():
    cnt = 0
    print('worker starts')
    time.sleep(3)
    while cnt < 100:
        print(cnt)
        cnt += 1
        time.sleep(1)


# ------------------------------------------------------------------------------
#
print(os.getpid())
dh = ru.DebugHelper()
t  = ru.Thread(name='worker', target=worker_outer)
t.start()
time.sleep(50)
t.join()




#!/usr/bin/env python

import zmq
import time
import msgpack

DELAY = 0.5


# ------------------------------------------------------------------------------
#
addr = None
with open('test.bridge.url', 'r') as fin:
    for line in fin.readlines():
        tag, addr = line.split()
        if tag == 'PUT':
            break

print('PUT: %s' % addr)

context    = zmq.Context()
socket     = context.socket(zmq.PUSH)
socket.hwm = 1
socket.connect(addr)

for n in range(1000):
    msg = {'data' : n}
    socket.send(msgpack.packb(msg))
    print('-> %s' % msg)
    time.sleep(DELAY)


# ------------------------------------------------------------------------------





import zmq

from typing import Any

import threading as mt

from ..json_io   import read_json
from ..misc      import as_string
from ..serialize import to_msgpack, from_msgpack
from ..logger    import Logger
from .utils      import no_intr, sock_connect


# ------------------------------------------------------------------------------
#
_LINGER_TIMEOUT    = 1000  # ms to linger after close
_HIGH_WATER_MARK   = 1024  # number of messages to buffer before dropping
_DEFAULT_BULK_SIZE =    1  # number of messages to put in a bulk


# ------------------------------------------------------------------------------
#
class Client(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, server: str    = None,
                       url:    str    = None,
                       log:    Logger = None) -> None:

        if server:
            self._url = read_json('%s.cfg' % server)['addr']

        elif url:
            self._url = url

        else:
            raise ValueError('need server name/cfg or Url')

        self._log  = log
        self._cb   = None
        self._ctx  = zmq.Context()
        self._sock = self._ctx.socket(zmq.REQ)

        self._sock.linger = _LINGER_TIMEOUT
        self._sock.hwm    = _HIGH_WATER_MARK

        sock_connect(self._sock, self._url)

        self._term   = mt.Event()
        self._active = False


    # --------------------------------------------------------------------------
    #
    @property
    def url(self) -> str:
        return self._url


    # --------------------------------------------------------------------------
    #
    def request(self, cmd: str, *args: Any, **kwargs: Any) -> Any:

        msg = {'cmd'   : cmd,
               'args'  : args,
               'kwargs': kwargs}

        if self._log:
            self._log.debug('request: %s', msg)

        req = to_msgpack(msg)

        no_intr(self._sock.send, req)

        msg = no_intr(self._sock.recv)
        res = as_string(from_msgpack(msg))

        # FIXME: assert proper res structure

        if res.get('err'):
            err_msg = 'ERROR: %s' % res['err']
            if res['exc']:
                err_msg += '\n%s' % ''.join(res['exc'])
            raise RuntimeError(err_msg)

        return res['res']


    # --------------------------------------------------------------------------
    #
    def close(self) -> None:

        self._sock.close()


# ------------------------------------------------------------------------------




#!/usr/bin/env python 

import sys
import zmq
import time

addr = None
with open('test.bridge.url', 'r') as fin:
    for line in fin.readlines():
        tag, addr = line.split()
        if tag == 'SUB':
            break

print('add: %s' % addr)

ctx = zmq.Context()
sub = ctx.socket(zmq.SUB)
sub.connect(addr)
sub.setsockopt_string(zmq.SUBSCRIBE, 'topic')

n = 0
start = time.time()
while True:
    top, msg = sub.recv_multipart()
    sys.stdout.write(msg)
    sys.stdout.flush()
    n += 1 
    if msg == 'x':
        break
print()
stop = time.time()
print('<- %.2f /s' % (n / (stop - start)))





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

import radical.utils  as ru
import does_not_exist as nope                 # noqa pylint: disable=F0401,W0611


# ------------------------------------------------------------------------------
#
PLUGIN_DESCRIPTION = {
    'type'       : 'unittests_2',
    'name'       : 'default_1',
    'class'      : 'PLUGIN_CLASS',
    'version'    : '0.1',
    'description': 'this is an empty test which basically does nothing.'
}


# ------------------------------------------------------------------------------
#
class PLUGIN_CLASS(ru.PluginBase, metaclass=ru.Singleton):
    '''
    This class implements the (empty) default unittest plugin for radical.utils.
    '''

    _created = False  # singleton test

    # --------------------------------------------------------------------------
    #
    def __init__(self, descr, *args, **kwargs):

        super().__init__(descr)

        if PLUGIN_CLASS._created:
            assert False, "singleton plugin created twice"

        PLUGIN_CLASS._created = True

        self._args   = args
        self._kwargs = kwargs


    # --------------------------------------------------------------------------
    #
    def run(self):

        return self._args


# ------------------------------------------------------------------------------






# ------------------------------------------------------------------------------
#
class SchedulerBase(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, resources):
        """
        Initialize the scheduler with a certain set and topology of
        resources to schedule requests over.
        """
        raise NotImplementedError('base class is virtual')


    # --------------------------------------------------------------------------
    #
    def alloc(self, req):
        """
        allocate a slice of the available resources to satisfy the request, and
        return that allocation.

        The returned allocation is a tuple [loc, req], where 'req' is the reqest
        passed as argument, and 'loc' is eithen one of the following:

          - int:          start index from the beginning of the resource list
          - list of ints: set of indexes in the resource list

        """
        raise NotImplementedError('base class is virtual')


    # --------------------------------------------------------------------------
    #
    def dealloc(self, res):
        """
        deallocate a given allocation to return its resources to the pool of
        resources available for new allocations.
        """
        raise NotImplementedError('base class is virtual')


    # --------------------------------------------------------------------------
    #
    def get_layout(self):
        """
        return row/col info for visual representation
        """
        raise NotImplementedError('base class is virtual')


    # --------------------------------------------------------------------------
    #
    def get_map(self):
        """
        return a map indicating current resource allocation
        """
        raise NotImplementedError('base class is virtual')


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import threading as mt

from .lockable  import Lockable
from .singleton import Singleton


# ------------------------------------------------------------------------------
#
# default timeout for delayed object removal.
#
_TIMEOUT    = 10
_DEFAULT_NS = 'global'


# ------------------------------------------------------------------------------
#
@Lockable
class ObjectCache(object, metaclass=Singleton):

    '''
    This is a singleton object caching class -- it maintains a reference
    counted registry of existing objects.
    '''

    # TODO: we should introduce namespaces -- this is a singleton, but we may
    #       want to use it in several places, thus need to make sure to not use
    #       colliding names...
    #
    # FIXME: this class is not thread-safe!


    # --------------------------------------------------------------------------
    #
    def __init__(self, timeout=_TIMEOUT):
        '''
        Make sure the object cache dict is initialized, exactly once.

        If timeout is 0 or smaller, the objects are removed immediately --
        otherwise removal is delayed by the specified timeout in seconds, to
        avoid thrashing on frequent removal/creation.
        '''

        self._timeout = timeout
        print(1, timeout, self._timeout)
        self._cache   = dict()


    # --------------------------------------------------------------------------
    #
    def get_obj(self, oid, creator, ns=_DEFAULT_NS):
        '''
        For a given object id, attempt to retrieve an existing object.  If that
        object exists, increase the reference counter, as there is now one more
        user for that object.

        If that object does not exist, call the given creator, then register and
        return the object thusly created.

        oid    : id of the object to get from the cache.
        creator: method to use to create a new object instance

                  Example:
                      def creator():
                          return Logger(name)

                      ret = object_cache.get_object(name, creator)
        '''

        with self:

            if ns not in self._cache:
                self._cache[ns] = dict()
            ns_cache = self._cache[ns]


            oid = str(oid)

            if oid not in ns_cache:

                obj = creator()

                ns_cache[oid]        = dict()
                ns_cache[oid]['cnt'] = 0
                ns_cache[oid]['obj'] = obj

            ns_cache[oid]['cnt'] += 1

            return ns_cache [oid]['obj']


    # --------------------------------------------------------------------------
    #
    def rem_obj(self, obj, ns=_DEFAULT_NS):
        '''
        For a given objects instance, decrease the refcounter as the caller
        stops using that object.  Once the ref counter is '0', remove all traces
        of the object -- this should make that object eligable for Python's
        garbage collection.  Returns 'True' if the given object was indeed
        registered, 'False' otherwise.

        The removal of the object is actually time-delayed.  That way, we will
        keep the object around *just* a little longer, which provides caching
        semantics in the case of frequent creation/dstruction cycles.
        '''

        with self:

            if ns not in self._cache:
                # impossible to find object
                return False

            ns_cache = self._cache[ns]

            for oid in ns_cache:

                if obj == ns_cache [oid]['obj']:

                    print(self._timeout)

                    if self._timeout:
                        # delay actual removeal by _timeout seconds
                        mt.Timer(self._timeout, self._rem_obj, [oid]).start()

                    else:
                        # immediate removeal
                        self._rem_obj(oid)

                    return True

            return False  # obj not found


    # --------------------------------------------------------------------------
    #
    def _rem_obj(self, oid, ns=_DEFAULT_NS):
        '''
        actual removal of an object(identified by oid) from the cache -- see
        :func:`rem_obj()` for details.
        '''

        with self:

            assert ns in self._cache

            ns_cache = self._cache[ns]
            ns_cache [oid]['cnt'] -= 1

            if ns_cache [oid]['cnt'] == 0:
                ns_cache [oid]['obj'] = None  # free the obj reference
                ns_cache.pop(oid, None)      # remove the cache entry


# ------------------------------------------------------------------------------




#!/usr/bin/env python

# ------------------------------------------------------------------------------
#
# watch a test for a specific condition to be met, and return the value of the
# condition (or None on timeout).
#
# ------------------------------------------------------------------------------

import radical.utils as ru
import time

start = time.time()


def test():
    check = bool(time.time() - start > 10)
    print(check)
    return check


print(ru.watch_condition(cond=test, target=True, timeout=2,    interval=1))
print(ru.watch_condition(cond=test, target=True, timeout=None, interval=1))


